# Licensed under the LGPL: https://www.gnu.org/licenses/old-licenses/lgpl-2.1.en.html
# For details: https://github.com/pylint-dev/astroid/blob/main/LICENSE
# Copyright (c) https://github.com/pylint-dev/astroid/blob/main/CONTRIBUTORS.txt

import textwrap

import pytest

import astroid
from astroid import builder, nodes
from astroid.const import PY310_PLUS, PY312_PLUS


class TestLinenoColOffset:
    """Test 'lineno', 'col_offset', 'end_lineno', and 'end_col_offset' for all
    nodes.
    """

    @staticmethod
    def test_end_lineno_container() -> None:
        """Container nodes: List, Tuple, Set."""
        code = textwrap.dedent(
            """
        [1, 2, 3]  #@
        [  #@
            1, 2, 3
        ]
        (1, 2, 3)  #@
        {1, 2, 3}  #@
        """
        ).strip()
        ast_nodes = builder.extract_node(code)
        assert isinstance(ast_nodes, list) and len(ast_nodes) == 4

        c1 = ast_nodes[0]
        assert isinstance(c1, nodes.List)
        assert (c1.lineno, c1.col_offset) == (1, 0)
        assert (c1.end_lineno, c1.end_col_offset) == (1, 9)

        c2 = ast_nodes[1]
        assert isinstance(c2, nodes.List)
        assert (c2.lineno, c2.col_offset) == (2, 0)
        assert (c2.end_lineno, c2.end_col_offset) == (4, 1)

        c3 = ast_nodes[2]
        assert isinstance(c3, nodes.Tuple)
        assert (c3.lineno, c3.col_offset) == (5, 0)
        assert (c3.end_lineno, c3.end_col_offset) == (5, 9)

        c4 = ast_nodes[3]
        assert isinstance(c4, nodes.Set)
        assert (c4.lineno, c4.col_offset) == (6, 0)
        assert (c4.end_lineno, c4.end_col_offset) == (6, 9)

    @staticmethod
    def test_end_lineno_name() -> None:
        """Name, Assign, AssignName, Delete, DelName."""
        code = textwrap.dedent(
            """
        var = 42  #@
        var  #@
        del var  #@

        var2 = (  #@
            1, 2, 3
        )
        """
        ).strip()
        ast_nodes = builder.extract_node(code)
        assert isinstance(ast_nodes, list) and len(ast_nodes) == 4

        n1 = ast_nodes[0]
        assert isinstance(n1, nodes.Assign)
        assert isinstance(n1.targets[0], nodes.AssignName)
        assert isinstance(n1.value, nodes.Const)
        assert (n1.lineno, n1.col_offset) == (1, 0)
        assert (n1.end_lineno, n1.end_col_offset) == (1, 8)
        assert (n1.targets[0].lineno, n1.targets[0].col_offset) == (1, 0)
        assert (n1.targets[0].end_lineno, n1.targets[0].end_col_offset) == (1, 3)
        assert (n1.value.lineno, n1.value.col_offset) == (1, 6)
        assert (n1.value.end_lineno, n1.value.end_col_offset) == (1, 8)

        n2 = ast_nodes[1]
        assert isinstance(n2, nodes.Name)
        assert (n2.lineno, n2.col_offset) == (2, 0)
        assert (n2.end_lineno, n2.end_col_offset) == (2, 3)

        n3 = ast_nodes[2]
        assert isinstance(n3, nodes.Delete) and isinstance(n3.targets[0], nodes.DelName)
        assert (n3.lineno, n3.col_offset) == (3, 0)
        assert (n3.end_lineno, n3.end_col_offset) == (3, 7)
        assert (n3.targets[0].lineno, n3.targets[0].col_offset) == (3, 4)
        assert (n3.targets[0].end_lineno, n3.targets[0].end_col_offset) == (3, 7)

        n4 = ast_nodes[3]
        assert isinstance(n4, nodes.Assign)
        assert isinstance(n4.targets[0], nodes.AssignName)
        assert (n4.lineno, n4.col_offset) == (5, 0)
        assert (n4.end_lineno, n4.end_col_offset) == (7, 1)
        assert (n4.targets[0].lineno, n4.targets[0].col_offset) == (5, 0)
        assert (n4.targets[0].end_lineno, n4.targets[0].end_col_offset) == (5, 4)

    @staticmethod
    def test_end_lineno_attribute() -> None:
        """Attribute, AssignAttr, DelAttr."""
        code = textwrap.dedent(
            """
        class X:
            var = 42

        X.var2 = 2  #@
        X.var2  #@
        del X.var2  #@
        """
        ).strip()
        ast_nodes = builder.extract_node(code)
        assert isinstance(ast_nodes, list) and len(ast_nodes) == 3

        a1 = ast_nodes[0]
        assert isinstance(a1, nodes.Assign)
        assert isinstance(a1.targets[0], nodes.AssignAttr)
        assert isinstance(a1.value, nodes.Const)
        assert (a1.lineno, a1.col_offset) == (4, 0)
        assert (a1.end_lineno, a1.end_col_offset) == (4, 10)
        assert (a1.targets[0].lineno, a1.targets[0].col_offset) == (4, 0)
        assert (a1.targets[0].end_lineno, a1.targets[0].end_col_offset) == (4, 6)
        assert (a1.value.lineno, a1.value.col_offset) == (4, 9)
        assert (a1.value.end_lineno, a1.value.end_col_offset) == (4, 10)

        a2 = ast_nodes[1]
        assert isinstance(a2, nodes.Attribute) and isinstance(a2.expr, nodes.Name)
        assert (a2.lineno, a2.col_offset) == (5, 0)
        assert (a2.end_lineno, a2.end_col_offset) == (5, 6)
        assert (a2.expr.lineno, a2.expr.col_offset) == (5, 0)
        assert (a2.expr.end_lineno, a2.expr.end_col_offset) == (5, 1)

        a3 = ast_nodes[2]
        assert isinstance(a3, nodes.Delete) and isinstance(a3.targets[0], nodes.DelAttr)
        assert (a3.lineno, a3.col_offset) == (6, 0)
        assert (a3.end_lineno, a3.end_col_offset) == (6, 10)
        assert (a3.targets[0].lineno, a3.targets[0].col_offset) == (6, 4)
        assert (a3.targets[0].end_lineno, a3.targets[0].end_col_offset) == (6, 10)

    @staticmethod
    def test_end_lineno_call() -> None:
        """Call, Keyword."""
        code = textwrap.dedent(
            """
        func(arg1, arg2=value)  #@
        """
        ).strip()
        c1 = builder.extract_node(code)
        assert isinstance(c1, nodes.Call)
        assert isinstance(c1.func, nodes.Name)
        assert isinstance(c1.args[0], nodes.Name)
        assert isinstance(c1.keywords[0], nodes.Keyword)
        assert isinstance(c1.keywords[0].value, nodes.Name)

        assert (c1.lineno, c1.col_offset) == (1, 0)
        assert (c1.end_lineno, c1.end_col_offset) == (1, 22)
        assert (c1.func.lineno, c1.func.col_offset) == (1, 0)
        assert (c1.func.end_lineno, c1.func.end_col_offset) == (1, 4)

        assert (c1.args[0].lineno, c1.args[0].col_offset) == (1, 5)
        assert (c1.args[0].end_lineno, c1.args[0].end_col_offset) == (1, 9)

        # fmt: off
        assert (c1.keywords[0].lineno, c1.keywords[0].col_offset) == (1, 11)
        assert (c1.keywords[0].end_lineno, c1.keywords[0].end_col_offset) == (1, 21)
        assert (c1.keywords[0].value.lineno, c1.keywords[0].value.col_offset) == (1, 16)
        assert (c1.keywords[0].value.end_lineno, c1.keywords[0].value.end_col_offset) == (1, 21)
        # fmt: on

    @staticmethod
    def test_end_lineno_assignment() -> None:
        """Assign, AnnAssign, AugAssign."""
        code = textwrap.dedent(
            """
        var = 2  #@
        var2: int = 2  #@
        var3 += 2  #@
        """
        ).strip()
        ast_nodes = builder.extract_node(code)
        assert isinstance(ast_nodes, list) and len(ast_nodes) == 3

        a1 = ast_nodes[0]
        assert isinstance(a1, nodes.Assign)
        assert isinstance(a1.targets[0], nodes.AssignName)
        assert isinstance(a1.value, nodes.Const)
        assert (a1.lineno, a1.col_offset) == (1, 0)
        assert (a1.end_lineno, a1.end_col_offset) == (1, 7)
        assert (a1.targets[0].lineno, a1.targets[0].col_offset) == (1, 0)
        assert (a1.targets[0].end_lineno, a1.targets[0].end_col_offset) == (1, 3)
        assert (a1.value.lineno, a1.value.col_offset) == (1, 6)
        assert (a1.value.end_lineno, a1.value.end_col_offset) == (1, 7)

        a2 = ast_nodes[1]
        assert isinstance(a2, nodes.AnnAssign)
        assert isinstance(a2.target, nodes.AssignName)
        assert isinstance(a2.annotation, nodes.Name)
        assert isinstance(a2.value, nodes.Const)
        assert (a2.lineno, a2.col_offset) == (2, 0)
        assert (a2.end_lineno, a2.end_col_offset) == (2, 13)
        assert (a2.target.lineno, a2.target.col_offset) == (2, 0)
        assert (a2.target.end_lineno, a2.target.end_col_offset) == (2, 4)
        assert (a2.annotation.lineno, a2.annotation.col_offset) == (2, 6)
        assert (a2.annotation.end_lineno, a2.annotation.end_col_offset) == (2, 9)
        assert (a2.value.lineno, a2.value.col_offset) == (2, 12)
        assert (a2.value.end_lineno, a2.value.end_col_offset) == (2, 13)

        a3 = ast_nodes[2]
        assert isinstance(a3, nodes.AugAssign)
        assert isinstance(a3.target, nodes.AssignName)
        assert isinstance(a3.value, nodes.Const)
        assert (a3.lineno, a3.col_offset) == (3, 0)
        assert (a3.end_lineno, a3.end_col_offset) == (3, 9)
        assert (a3.target.lineno, a3.target.col_offset) == (3, 0)
        assert (a3.target.end_lineno, a3.target.end_col_offset) == (3, 4)
        assert (a3.value.lineno, a3.value.col_offset) == (3, 8)
        assert (a3.value.end_lineno, a3.value.end_col_offset) == (3, 9)

    @staticmethod
    def test_end_lineno_mix_stmts() -> None:
        """Assert, Break, Continue, Global, Nonlocal, Pass, Raise, Return, Expr."""
        code = textwrap.dedent(
            """
        assert True, "Some message"  #@
        break  #@
        continue  #@
        global var  #@
        nonlocal var  #@
        pass  #@
        raise Exception from ex  #@
        return 42  #@
        var  #@
        """
        ).strip()
        ast_nodes = builder.extract_node(code)
        assert isinstance(ast_nodes, list) and len(ast_nodes) == 9

        s1 = ast_nodes[0]
        assert isinstance(s1, nodes.Assert)
        assert isinstance(s1.test, nodes.Const)
        assert isinstance(s1.fail, nodes.Const)
        assert (s1.lineno, s1.col_offset) == (1, 0)
        assert (s1.end_lineno, s1.end_col_offset) == (1, 27)
        assert (s1.test.lineno, s1.test.col_offset) == (1, 7)
        assert (s1.test.end_lineno, s1.test.end_col_offset) == (1, 11)
        assert (s1.fail.lineno, s1.fail.col_offset) == (1, 13)
        assert (s1.fail.end_lineno, s1.fail.end_col_offset) == (1, 27)

        s2 = ast_nodes[1]
        assert isinstance(s2, nodes.Break)
        assert (s2.lineno, s2.col_offset) == (2, 0)
        assert (s2.end_lineno, s2.end_col_offset) == (2, 5)

        s3 = ast_nodes[2]
        assert isinstance(s3, nodes.Continue)
        assert (s3.lineno, s3.col_offset) == (3, 0)
        assert (s3.end_lineno, s3.end_col_offset) == (3, 8)

        s4 = ast_nodes[3]
        assert isinstance(s4, nodes.Global)
        assert (s4.lineno, s4.col_offset) == (4, 0)
        assert (s4.end_lineno, s4.end_col_offset) == (4, 10)

        s5 = ast_nodes[4]
        assert isinstance(s5, nodes.Nonlocal)
        assert (s5.lineno, s5.col_offset) == (5, 0)
        assert (s5.end_lineno, s5.end_col_offset) == (5, 12)

        s6 = ast_nodes[5]
        assert isinstance(s6, nodes.Pass)
        assert (s6.lineno, s6.col_offset) == (6, 0)
        assert (s6.end_lineno, s6.end_col_offset) == (6, 4)

        s7 = ast_nodes[6]
        assert isinstance(s7, nodes.Raise)
        assert isinstance(s7.exc, nodes.Name)
        assert isinstance(s7.cause, nodes.Name)
        assert (s7.lineno, s7.col_offset) == (7, 0)
        assert (s7.end_lineno, s7.end_col_offset) == (7, 23)
        assert (s7.exc.lineno, s7.exc.col_offset) == (7, 6)
        assert (s7.exc.end_lineno, s7.exc.end_col_offset) == (7, 15)
        assert (s7.cause.lineno, s7.cause.col_offset) == (7, 21)
        assert (s7.cause.end_lineno, s7.cause.end_col_offset) == (7, 23)

        s8 = ast_nodes[7]
        assert isinstance(s8, nodes.Return)
        assert isinstance(s8.value, nodes.Const)
        assert (s8.lineno, s8.col_offset) == (8, 0)
        assert (s8.end_lineno, s8.end_col_offset) == (8, 9)
        assert (s8.value.lineno, s8.value.col_offset) == (8, 7)
        assert (s8.value.end_lineno, s8.value.end_col_offset) == (8, 9)

        s9 = ast_nodes[8].parent
        assert isinstance(s9, nodes.Expr)
        assert isinstance(s9.value, nodes.Name)
        assert (s9.lineno, s9.col_offset) == (9, 0)
        assert (s9.end_lineno, s9.end_col_offset) == (9, 3)
        assert (s9.value.lineno, s9.value.col_offset) == (9, 0)
        assert (s9.value.end_lineno, s9.value.end_col_offset) == (9, 3)

    @staticmethod
    def test_end_lineno_mix_nodes() -> None:
        """Await, Starred, Yield, YieldFrom."""
        code = textwrap.dedent(
            """
        await func  #@
        *args  #@
        yield 42  #@
        yield from (1, 2)  #@
        """
        ).strip()
        ast_nodes = builder.extract_node(code)
        assert isinstance(ast_nodes, list) and len(ast_nodes) == 4

        n1 = ast_nodes[0]
        assert isinstance(n1, nodes.Await)
        assert isinstance(n1.value, nodes.Name)
        assert (n1.lineno, n1.col_offset) == (1, 0)
        assert (n1.end_lineno, n1.end_col_offset) == (1, 10)
        assert (n1.value.lineno, n1.value.col_offset) == (1, 6)
        assert (n1.value.end_lineno, n1.value.end_col_offset) == (1, 10)

        n2 = ast_nodes[1]
        assert isinstance(n2, nodes.Starred)
        assert isinstance(n2.value, nodes.Name)
        assert (n2.lineno, n2.col_offset) == (2, 0)
        assert (n2.end_lineno, n2.end_col_offset) == (2, 5)
        assert (n2.value.lineno, n2.value.col_offset) == (2, 1)
        assert (n2.value.end_lineno, n2.value.end_col_offset) == (2, 5)

        n3 = ast_nodes[2]
        assert isinstance(n3, nodes.Yield)
        assert isinstance(n3.value, nodes.Const)
        assert (n3.lineno, n3.col_offset) == (3, 0)
        assert (n3.end_lineno, n3.end_col_offset) == (3, 8)
        assert (n3.value.lineno, n3.value.col_offset) == (3, 6)
        assert (n3.value.end_lineno, n3.value.end_col_offset) == (3, 8)

        n4 = ast_nodes[3]
        assert isinstance(n4, nodes.YieldFrom)
        assert isinstance(n4.value, nodes.Tuple)
        assert (n4.lineno, n4.col_offset) == (4, 0)
        assert (n4.end_lineno, n4.end_col_offset) == (4, 17)
        assert (n4.value.lineno, n4.value.col_offset) == (4, 11)
        assert (n4.value.end_lineno, n4.value.end_col_offset) == (4, 17)

    @staticmethod
    def test_end_lineno_ops() -> None:
        """BinOp, BoolOp, UnaryOp, Compare."""
        code = textwrap.dedent(
            """
        x + y  #@
        a and b  #@
        -var  #@
        a < b  #@
        """
        ).strip()
        ast_nodes = builder.extract_node(code)
        assert isinstance(ast_nodes, list) and len(ast_nodes) == 4

        o1 = ast_nodes[0]
        assert isinstance(o1, nodes.BinOp)
        assert isinstance(o1.left, nodes.Name)
        assert isinstance(o1.right, nodes.Name)
        assert (o1.lineno, o1.col_offset) == (1, 0)
        assert (o1.end_lineno, o1.end_col_offset) == (1, 5)
        assert (o1.left.lineno, o1.left.col_offset) == (1, 0)
        assert (o1.left.end_lineno, o1.left.end_col_offset) == (1, 1)
        assert (o1.right.lineno, o1.right.col_offset) == (1, 4)
        assert (o1.right.end_lineno, o1.right.end_col_offset) == (1, 5)

        o2 = ast_nodes[1]
        assert isinstance(o2, nodes.BoolOp)
        assert isinstance(o2.values[0], nodes.Name)
        assert isinstance(o2.values[1], nodes.Name)
        assert (o2.lineno, o2.col_offset) == (2, 0)
        assert (o2.end_lineno, o2.end_col_offset) == (2, 7)
        assert (o2.values[0].lineno, o2.values[0].col_offset) == (2, 0)
        assert (o2.values[0].end_lineno, o2.values[0].end_col_offset) == (2, 1)
        assert (o2.values[1].lineno, o2.values[1].col_offset) == (2, 6)
        assert (o2.values[1].end_lineno, o2.values[1].end_col_offset) == (2, 7)

        o3 = ast_nodes[2]
        assert isinstance(o3, nodes.UnaryOp)
        assert isinstance(o3.operand, nodes.Name)
        assert (o3.lineno, o3.col_offset) == (3, 0)
        assert (o3.end_lineno, o3.end_col_offset) == (3, 4)
        assert (o3.operand.lineno, o3.operand.col_offset) == (3, 1)
        assert (o3.operand.end_lineno, o3.operand.end_col_offset) == (3, 4)

        o4 = ast_nodes[3]
        assert isinstance(o4, nodes.Compare)
        assert isinstance(o4.left, nodes.Name)
        assert isinstance(o4.ops[0][1], nodes.Name)
        assert (o4.lineno, o4.col_offset) == (4, 0)
        assert (o4.end_lineno, o4.end_col_offset) == (4, 5)
        assert (o4.left.lineno, o4.left.col_offset) == (4, 0)
        assert (o4.left.end_lineno, o4.left.end_col_offset) == (4, 1)
        assert (o4.ops[0][1].lineno, o4.ops[0][1].col_offset) == (4, 4)
        assert (o4.ops[0][1].end_lineno, o4.ops[0][1].end_col_offset) == (4, 5)

    @staticmethod
    def test_end_lineno_if() -> None:
        """If, IfExp, NamedExpr."""
        code = textwrap.dedent(
            """
        if (  #@
            var := 2  #@
        ):
            pass
        else:
            pass

        2 if True else 1  #@
        """
        ).strip()
        ast_nodes = builder.extract_node(code)
        assert isinstance(ast_nodes, list) and len(ast_nodes) == 3

        i1 = ast_nodes[0]
        assert isinstance(i1, nodes.If)
        assert isinstance(i1.test, nodes.NamedExpr)
        assert isinstance(i1.body[0], nodes.Pass)
        assert isinstance(i1.orelse[0], nodes.Pass)
        assert (i1.lineno, i1.col_offset) == (1, 0)
        assert (i1.end_lineno, i1.end_col_offset) == (6, 8)
        assert (i1.test.lineno, i1.test.col_offset) == (2, 4)
        assert (i1.test.end_lineno, i1.test.end_col_offset) == (2, 12)
        assert (i1.body[0].lineno, i1.body[0].col_offset) == (4, 4)
        assert (i1.body[0].end_lineno, i1.body[0].end_col_offset) == (4, 8)
        assert (i1.orelse[0].lineno, i1.orelse[0].col_offset) == (6, 4)
        assert (i1.orelse[0].end_lineno, i1.orelse[0].end_col_offset) == (6, 8)

        i2 = ast_nodes[1]
        assert isinstance(i2, nodes.NamedExpr)
        assert isinstance(i2.target, nodes.AssignName)
        assert isinstance(i2.value, nodes.Const)
        assert (i2.lineno, i2.col_offset) == (2, 4)
        assert (i2.end_lineno, i2.end_col_offset) == (2, 12)
        assert (i2.target.lineno, i2.target.col_offset) == (2, 4)
        assert (i2.target.end_lineno, i2.target.end_col_offset) == (2, 7)
        assert (i2.value.lineno, i2.value.col_offset) == (2, 11)
        assert (i2.value.end_lineno, i2.value.end_col_offset) == (2, 12)

        i3 = ast_nodes[2]
        assert isinstance(i3, nodes.IfExp)
        assert isinstance(i3.test, nodes.Const)
        assert isinstance(i3.body, nodes.Const)
        assert isinstance(i3.orelse, nodes.Const)
        assert (i3.lineno, i3.col_offset) == (8, 0)
        assert (i3.end_lineno, i3.end_col_offset) == (8, 16)
        assert (i3.test.lineno, i3.test.col_offset) == (8, 5)
        assert (i3.test.end_lineno, i3.test.end_col_offset) == (8, 9)
        assert (i3.body.lineno, i3.body.col_offset) == (8, 0)
        assert (i3.body.end_lineno, i3.body.end_col_offset) == (8, 1)
        assert (i3.orelse.lineno, i3.orelse.col_offset) == (8, 15)
        assert (i3.orelse.end_lineno, i3.orelse.end_col_offset) == (8, 16)

    @staticmethod
    def test_end_lineno_for() -> None:
        """For, AsyncFor."""
        code = textwrap.dedent(
            """
        for i in lst:  #@
            pass
        else:
            pass

        async for i in lst:  #@
            pass
        """
        ).strip()
        ast_nodes = builder.extract_node(code)
        assert isinstance(ast_nodes, list) and len(ast_nodes) == 2

        f1 = ast_nodes[0]
        assert isinstance(f1, nodes.For)
        assert isinstance(f1.target, nodes.AssignName)
        assert isinstance(f1.iter, nodes.Name)
        assert isinstance(f1.body[0], nodes.Pass)
        assert isinstance(f1.orelse[0], nodes.Pass)
        assert (f1.lineno, f1.col_offset) == (1, 0)
        assert (f1.end_lineno, f1.end_col_offset) == (4, 8)
        assert (f1.target.lineno, f1.target.col_offset) == (1, 4)
        assert (f1.target.end_lineno, f1.target.end_col_offset) == (1, 5)
        assert (f1.iter.lineno, f1.iter.col_offset) == (1, 9)
        assert (f1.iter.end_lineno, f1.iter.end_col_offset) == (1, 12)
        assert (f1.body[0].lineno, f1.body[0].col_offset) == (2, 4)
        assert (f1.body[0].end_lineno, f1.body[0].end_col_offset) == (2, 8)
        assert (f1.orelse[0].lineno, f1.orelse[0].col_offset) == (4, 4)
        assert (f1.orelse[0].end_lineno, f1.orelse[0].end_col_offset) == (4, 8)

        f2 = ast_nodes[1]
        assert isinstance(f2, nodes.AsyncFor)
        assert isinstance(f2.target, nodes.AssignName)
        assert isinstance(f2.iter, nodes.Name)
        assert isinstance(f2.body[0], nodes.Pass)
        assert (f2.lineno, f2.col_offset) == (6, 0)
        assert (f2.end_lineno, f2.end_col_offset) == (7, 8)
        assert (f2.target.lineno, f2.target.col_offset) == (6, 10)
        assert (f2.target.end_lineno, f2.target.end_col_offset) == (6, 11)
        assert (f2.iter.lineno, f2.iter.col_offset) == (6, 15)
        assert (f2.iter.end_lineno, f2.iter.end_col_offset) == (6, 18)
        assert (f2.body[0].lineno, f2.body[0].col_offset) == (7, 4)
        assert (f2.body[0].end_lineno, f2.body[0].end_col_offset) == (7, 8)

    @staticmethod
    def test_end_lineno_const() -> None:
        """Const (int, str, bool, None, bytes, ellipsis)."""
        code = textwrap.dedent(
            """
        2  #@
        "Hello"  #@
        True  #@
        None  #@
        b"01"  #@
        ...  #@
        """
        ).strip()
        ast_nodes = builder.extract_node(code)
        assert isinstance(ast_nodes, list) and len(ast_nodes) == 6

        c1 = ast_nodes[0]
        assert isinstance(c1, nodes.Const)
        assert (c1.lineno, c1.col_offset) == (1, 0)
        assert (c1.end_lineno, c1.end_col_offset) == (1, 1)

        c2 = ast_nodes[1]
        assert isinstance(c2, nodes.Const)
        assert (c2.lineno, c2.col_offset) == (2, 0)
        assert (c2.end_lineno, c2.end_col_offset) == (2, 7)

        c3 = ast_nodes[2]
        assert isinstance(c3, nodes.Const)
        assert (c3.lineno, c3.col_offset) == (3, 0)
        assert (c3.end_lineno, c3.end_col_offset) == (3, 4)

        c4 = ast_nodes[3]
        assert isinstance(c4, nodes.Const)
        assert (c4.lineno, c4.col_offset) == (4, 0)
        assert (c4.end_lineno, c4.end_col_offset) == (4, 4)

        c5 = ast_nodes[4]
        assert isinstance(c5, nodes.Const)
        assert (c5.lineno, c5.col_offset) == (5, 0)
        assert (c5.end_lineno, c5.end_col_offset) == (5, 5)

        c6 = ast_nodes[5]
        assert isinstance(c6, nodes.Const)
        assert (c6.lineno, c6.col_offset) == (6, 0)
        assert (c6.end_lineno, c6.end_col_offset) == (6, 3)

    @staticmethod
    def test_end_lineno_function() -> None:
        """FunctionDef, AsyncFunctionDef, Decorators, Lambda, Arguments."""
        code = textwrap.dedent(
            """
        def func(  #@
            a: int = 0, /,
            var: int = 1, *args: Any,
            keyword: int = 2, **kwargs: Any
        ) -> None:
            pass

        @decorator1
        @decorator2
        async def func():  #@
            pass

        lambda x: 2  #@
        """
        ).strip()
        ast_nodes = builder.extract_node(code)
        assert isinstance(ast_nodes, list) and len(ast_nodes) == 3

        # fmt: off
        f1 = ast_nodes[0]
        assert isinstance(f1, nodes.FunctionDef)
        assert isinstance(f1.args, nodes.Arguments)
        assert isinstance(f1.returns, nodes.Const)
        assert isinstance(f1.body[0], nodes.Pass)
        assert (f1.lineno, f1.col_offset) == (1, 0)
        assert (f1.end_lineno, f1.end_col_offset) == (6, 8)
        assert (f1.returns.lineno, f1.returns.col_offset) == (5, 5)
        assert (f1.returns.end_lineno, f1.returns.end_col_offset) == (5, 9)
        assert (f1.body[0].lineno, f1.body[0].col_offset) == (6, 4)
        assert (f1.body[0].end_lineno, f1.body[0].end_col_offset) == (6, 8)

        # pos only arguments
        # TODO fix column offset: arg -> arg (AssignName)
        assert isinstance(f1.args.posonlyargs[0], nodes.AssignName)
        assert (f1.args.posonlyargs[0].lineno, f1.args.posonlyargs[0].col_offset) == (2, 4)
        assert (f1.args.posonlyargs[0].end_lineno, f1.args.posonlyargs[0].end_col_offset) == (2, 10)
        assert isinstance(f1.args.posonlyargs_annotations[0], nodes.Name)
        assert (
            f1.args.posonlyargs_annotations[0].lineno, f1.args.posonlyargs_annotations[0].col_offset
        ) == (2, 7)
        assert (
            f1.args.posonlyargs_annotations[0].end_lineno, f1.args.posonlyargs_annotations[0].end_col_offset
        ) == (2, 10)
        assert (f1.args.defaults[0].lineno, f1.args.defaults[0].col_offset) == (2, 13)
        assert (f1.args.defaults[0].end_lineno, f1.args.defaults[0].end_col_offset) == (2, 14)

        # pos or kw arguments
        assert isinstance(f1.args.args[0], nodes.AssignName)
        assert (f1.args.args[0].lineno, f1.args.args[0].col_offset) == (3, 4)
        assert (f1.args.args[0].end_lineno, f1.args.args[0].end_col_offset) == (3, 12)
        assert isinstance(f1.args.annotations[0], nodes.Name)
        assert (f1.args.annotations[0].lineno, f1.args.annotations[0].col_offset) == (3, 9)
        assert (f1.args.annotations[0].end_lineno, f1.args.annotations[0].end_col_offset) == (3, 12)
        assert isinstance(f1.args.defaults[1], nodes.Const)
        assert (f1.args.defaults[1].lineno, f1.args.defaults[1].col_offset) == (3, 15)
        assert (f1.args.defaults[1].end_lineno, f1.args.defaults[1].end_col_offset) == (3, 16)

        # *args
        assert isinstance(f1.args.varargannotation, nodes.Name)
        assert (f1.args.varargannotation.lineno, f1.args.varargannotation.col_offset) == (3, 25)
        assert (f1.args.varargannotation.end_lineno, f1.args.varargannotation.end_col_offset) == (3, 28)

        # kw_only arguments
        assert isinstance(f1.args.kwonlyargs[0], nodes.AssignName)
        assert (f1.args.kwonlyargs[0].lineno, f1.args.kwonlyargs[0].col_offset) == (4, 4)
        assert (f1.args.kwonlyargs[0].end_lineno, f1.args.kwonlyargs[0].end_col_offset) == (4, 16)
        annotations = f1.args.kwonlyargs_annotations
        assert isinstance(annotations[0], nodes.Name)
        assert (annotations[0].lineno, annotations[0].col_offset) == (4, 13)
        assert (annotations[0].end_lineno, annotations[0].end_col_offset) == (4, 16)
        assert isinstance(f1.args.kw_defaults[0], nodes.Const)
        assert (f1.args.kw_defaults[0].lineno, f1.args.kw_defaults[0].col_offset) == (4, 19)
        assert (f1.args.kw_defaults[0].end_lineno, f1.args.kw_defaults[0].end_col_offset) == (4, 20)

        # **kwargs
        assert isinstance(f1.args.kwargannotation, nodes.Name)
        assert (f1.args.kwargannotation.lineno, f1.args.kwargannotation.col_offset) == (4, 32)
        assert (f1.args.kwargannotation.end_lineno, f1.args.kwargannotation.end_col_offset) == (4, 35)

        f2 = ast_nodes[1]
        assert isinstance(f2, nodes.AsyncFunctionDef)
        assert isinstance(f2.decorators, nodes.Decorators)
        assert isinstance(f2.decorators.nodes[0], nodes.Name)
        assert isinstance(f2.decorators.nodes[1], nodes.Name)
        assert (f2.lineno, f2.col_offset) == (8, 0)
        assert (f2.end_lineno, f2.end_col_offset) == (11, 8)
        assert (f2.decorators.lineno, f2.decorators.col_offset) == (8, 0)
        assert (f2.decorators.end_lineno, f2.decorators.end_col_offset) == (9, 11)
        assert (f2.decorators.nodes[0].lineno, f2.decorators.nodes[0].col_offset) == (8, 1)
        assert (f2.decorators.nodes[0].end_lineno, f2.decorators.nodes[0].end_col_offset) == (8, 11)
        assert (f2.decorators.nodes[1].lineno, f2.decorators.nodes[1].col_offset) == (9, 1)
        assert (f2.decorators.nodes[1].end_lineno, f2.decorators.nodes[1].end_col_offset) == (9, 11)

        f3 = ast_nodes[2]
        assert isinstance(f3, nodes.Lambda)
        assert isinstance(f3.args, nodes.Arguments)
        assert isinstance(f3.args.args[0], nodes.AssignName)
        assert isinstance(f3.body, nodes.Const)
        assert (f3.lineno, f3.col_offset) == (13, 0)
        assert (f3.end_lineno, f3.end_col_offset) == (13, 11)
        assert (f3.args.args[0].lineno, f3.args.args[0].col_offset) == (13, 7)
        assert (f3.args.args[0].end_lineno, f3.args.args[0].end_col_offset) == (13, 8)
        assert (f3.body.lineno, f3.body.col_offset) == (13, 10)
        assert (f3.body.end_lineno, f3.body.end_col_offset) == (13, 11)
        # fmt: on

    @staticmethod
    def test_end_lineno_dict() -> None:
        """Dict, DictUnpack."""
        code = textwrap.dedent(
            """
        {  #@
            1: "Hello",
            **{2: "World"}  #@
        }
        """
        ).strip()
        ast_nodes = builder.extract_node(code)
        assert isinstance(ast_nodes, list) and len(ast_nodes) == 2

        d1 = ast_nodes[0]
        assert isinstance(d1, nodes.Dict)
        assert isinstance(d1.items[0][0], nodes.Const)
        assert isinstance(d1.items[0][1], nodes.Const)
        assert (d1.lineno, d1.col_offset) == (1, 0)
        assert (d1.end_lineno, d1.end_col_offset) == (4, 1)
        assert (d1.items[0][0].lineno, d1.items[0][0].col_offset) == (2, 4)
        assert (d1.items[0][0].end_lineno, d1.items[0][0].end_col_offset) == (2, 5)
        assert (d1.items[0][1].lineno, d1.items[0][1].col_offset) == (2, 7)
        assert (d1.items[0][1].end_lineno, d1.items[0][1].end_col_offset) == (2, 14)

        d2 = ast_nodes[1]
        assert isinstance(d2, nodes.DictUnpack)
        assert (d2.lineno, d2.col_offset) == (3, 6)
        assert (d2.end_lineno, d2.end_col_offset) == (3, 18)

    @staticmethod
    def test_end_lineno_try() -> None:
        """Try, ExceptHandler."""
        code = textwrap.dedent(
            """
        try:  #@
            pass
        except KeyError as ex:
            pass
        except AttributeError as ex:
            pass
        else:
            pass

        try:  #@
            pass
        except KeyError as ex:
            pass
        else:
            pass
        finally:
            pass
        """
        ).strip()
        ast_nodes = builder.extract_node(code)
        assert isinstance(ast_nodes, list) and len(ast_nodes) == 2

        t1 = ast_nodes[0]
        assert isinstance(t1, nodes.Try)
        assert isinstance(t1.body[0], nodes.Pass)
        assert isinstance(t1.orelse[0], nodes.Pass)
        assert (t1.lineno, t1.col_offset) == (1, 0)
        assert (t1.end_lineno, t1.end_col_offset) == (8, 8)
        assert (t1.body[0].lineno, t1.body[0].col_offset) == (2, 4)
        assert (t1.body[0].end_lineno, t1.body[0].end_col_offset) == (2, 8)
        assert (t1.orelse[0].lineno, t1.orelse[0].col_offset) == (8, 4)
        assert (t1.orelse[0].end_lineno, t1.orelse[0].end_col_offset) == (8, 8)

        t2 = t1.handlers[0]
        assert isinstance(t2, nodes.ExceptHandler)
        assert isinstance(t2.type, nodes.Name)
        assert isinstance(t2.name, nodes.AssignName)
        assert isinstance(t2.body[0], nodes.Pass)
        assert (t2.lineno, t2.col_offset) == (3, 0)
        assert (t2.end_lineno, t2.end_col_offset) == (4, 8)
        assert (t2.type.lineno, t2.type.col_offset) == (3, 7)
        assert (t2.type.end_lineno, t2.type.end_col_offset) == (3, 15)
        # TODO fix column offset: ExceptHandler -> name (AssignName)
        assert (t2.name.lineno, t2.name.col_offset) == (3, 0)
        assert (t2.name.end_lineno, t2.name.end_col_offset) == (4, 8)
        assert (t2.body[0].lineno, t2.body[0].col_offset) == (4, 4)
        assert (t2.body[0].end_lineno, t2.body[0].end_col_offset) == (4, 8)

        t3 = ast_nodes[1]
        assert isinstance(t3, nodes.Try)
        assert isinstance(t3.finalbody[0], nodes.Pass)
        assert (t3.lineno, t3.col_offset) == (10, 0)
        assert (t3.end_lineno, t3.end_col_offset) == (17, 8)
        assert (t3.body[0].lineno, t3.body[0].col_offset) == (11, 4)
        assert (t3.body[0].end_lineno, t3.body[0].end_col_offset) == (11, 8)
        assert (t3.finalbody[0].lineno, t3.finalbody[0].col_offset) == (17, 4)
        assert (t3.finalbody[0].end_lineno, t3.finalbody[0].end_col_offset) == (17, 8)

    @staticmethod
    def test_end_lineno_subscript() -> None:
        """Subscript, Slice, (ExtSlice, Index)."""
        code = textwrap.dedent(
            """
        var[0]  #@
        var[1:2:1]  #@
        var[1:2, 2]  #@
        """
        ).strip()
        ast_nodes = builder.extract_node(code)
        assert isinstance(ast_nodes, list) and len(ast_nodes) == 3

        s1 = ast_nodes[0]
        assert isinstance(s1, nodes.Subscript)
        assert isinstance(s1.value, nodes.Name)
        assert isinstance(s1.slice, nodes.Const)
        assert (s1.lineno, s1.col_offset) == (1, 0)
        assert (s1.end_lineno, s1.end_col_offset) == (1, 6)
        assert (s1.value.lineno, s1.value.col_offset) == (1, 0)
        assert (s1.value.end_lineno, s1.value.end_col_offset) == (1, 3)
        assert (s1.slice.lineno, s1.slice.col_offset) == (1, 4)
        assert (s1.slice.end_lineno, s1.slice.end_col_offset) == (1, 5)

        s2 = ast_nodes[1]
        assert isinstance(s2, nodes.Subscript)
        assert isinstance(s2.slice, nodes.Slice)
        assert isinstance(s2.slice.lower, nodes.Const)
        assert isinstance(s2.slice.upper, nodes.Const)
        assert isinstance(s2.slice.step, nodes.Const)
        assert (s2.lineno, s2.col_offset) == (2, 0)
        assert (s2.end_lineno, s2.end_col_offset) == (2, 10)
        assert (s2.slice.lower.lineno, s2.slice.lower.col_offset) == (2, 4)
        assert (s2.slice.lower.end_lineno, s2.slice.lower.end_col_offset) == (2, 5)
        assert (s2.slice.upper.lineno, s2.slice.upper.col_offset) == (2, 6)
        assert (s2.slice.upper.end_lineno, s2.slice.upper.end_col_offset) == (2, 7)
        assert (s2.slice.step.lineno, s2.slice.step.col_offset) == (2, 8)
        assert (s2.slice.step.end_lineno, s2.slice.step.end_col_offset) == (2, 9)

        s3 = ast_nodes[2]
        assert isinstance(s3, nodes.Subscript)
        assert isinstance(s3.slice, nodes.Tuple)
        assert (s3.lineno, s3.col_offset) == (3, 0)
        assert (s3.end_lineno, s3.end_col_offset) == (3, 11)
        assert (s3.slice.lineno, s3.slice.col_offset) == (3, 4)
        assert (s3.slice.end_lineno, s3.slice.end_col_offset) == (3, 10)

    @staticmethod
    def test_end_lineno_import() -> None:
        """Import, ImportFrom."""
        code = textwrap.dedent(
            """
        import a.b  #@
        import a as x  #@
        from . import x  #@
        from .a import y as y  #@
        """
        ).strip()
        ast_nodes = builder.extract_node(code)
        assert isinstance(ast_nodes, list) and len(ast_nodes) == 4

        i1 = ast_nodes[0]
        assert isinstance(i1, nodes.Import)
        assert (i1.lineno, i1.col_offset) == (1, 0)
        assert (i1.end_lineno, i1.end_col_offset) == (1, 10)

        i2 = ast_nodes[1]
        assert isinstance(i2, nodes.Import)
        assert (i2.lineno, i2.col_offset) == (2, 0)
        assert (i2.end_lineno, i2.end_col_offset) == (2, 13)

        i3 = ast_nodes[2]
        assert isinstance(i3, nodes.ImportFrom)
        assert (i3.lineno, i3.col_offset) == (3, 0)
        assert (i3.end_lineno, i3.end_col_offset) == (3, 15)

        i4 = ast_nodes[3]
        assert isinstance(i4, nodes.ImportFrom)
        assert (i4.lineno, i4.col_offset) == (4, 0)
        assert (i4.end_lineno, i4.end_col_offset) == (4, 21)

    @staticmethod
    def test_end_lineno_with() -> None:
        """With, AsyncWith."""
        code = textwrap.dedent(
            """
        with open(file) as fp, \\
                open(file2) as fp2:  #@
            pass

        async with open(file) as fp:  #@
            pass
        """
        ).strip()
        ast_nodes = builder.extract_node(code)
        assert isinstance(ast_nodes, list) and len(ast_nodes) == 2

        w1 = ast_nodes[0].parent
        assert isinstance(w1, nodes.With)
        assert isinstance(w1.items[0][0], nodes.Call)
        assert isinstance(w1.items[0][1], nodes.AssignName)
        assert isinstance(w1.items[1][0], nodes.Call)
        assert isinstance(w1.items[1][1], nodes.AssignName)
        assert isinstance(w1.body[0], nodes.Pass)
        assert (w1.lineno, w1.col_offset) == (1, 0)
        assert (w1.end_lineno, w1.end_col_offset) == (3, 8)
        assert (w1.items[0][0].lineno, w1.items[0][0].col_offset) == (1, 5)
        assert (w1.items[0][0].end_lineno, w1.items[0][0].end_col_offset) == (1, 15)
        assert (w1.items[0][1].lineno, w1.items[0][1].col_offset) == (1, 19)
        assert (w1.items[0][1].end_lineno, w1.items[0][1].end_col_offset) == (1, 21)
        assert (w1.items[1][0].lineno, w1.items[1][0].col_offset) == (2, 8)
        assert (w1.items[1][0].end_lineno, w1.items[1][0].end_col_offset) == (2, 19)
        assert (w1.items[1][1].lineno, w1.items[1][1].col_offset) == (2, 23)
        assert (w1.items[1][1].end_lineno, w1.items[1][1].end_col_offset) == (2, 26)
        assert (w1.body[0].lineno, w1.body[0].col_offset) == (3, 4)
        assert (w1.body[0].end_lineno, w1.body[0].end_col_offset) == (3, 8)

        w2 = ast_nodes[1]
        assert isinstance(w2, nodes.AsyncWith)
        assert isinstance(w2.items[0][0], nodes.Call)
        assert isinstance(w2.items[0][1], nodes.AssignName)
        assert isinstance(w2.body[0], nodes.Pass)
        assert (w2.lineno, w2.col_offset) == (5, 0)
        assert (w2.end_lineno, w2.end_col_offset) == (6, 8)
        assert (w2.items[0][0].lineno, w2.items[0][0].col_offset) == (5, 11)
        assert (w2.items[0][0].end_lineno, w2.items[0][0].end_col_offset) == (5, 21)
        assert (w2.items[0][1].lineno, w2.items[0][1].col_offset) == (5, 25)
        assert (w2.items[0][1].end_lineno, w2.items[0][1].end_col_offset) == (5, 27)
        assert (w2.body[0].lineno, w2.body[0].col_offset) == (6, 4)
        assert (w2.body[0].end_lineno, w2.body[0].end_col_offset) == (6, 8)

    @staticmethod
    def test_end_lineno_while() -> None:
        """While."""
        code = textwrap.dedent(
            """
        while 2:
            pass
        else:
            pass
        """
        ).strip()
        w1 = builder.extract_node(code)
        assert isinstance(w1, nodes.While)
        assert isinstance(w1.test, nodes.Const)
        assert isinstance(w1.body[0], nodes.Pass)
        assert isinstance(w1.orelse[0], nodes.Pass)
        assert (w1.lineno, w1.col_offset) == (1, 0)
        assert (w1.end_lineno, w1.end_col_offset) == (4, 8)
        assert (w1.test.lineno, w1.test.col_offset) == (1, 6)
        assert (w1.test.end_lineno, w1.test.end_col_offset) == (1, 7)
        assert (w1.body[0].lineno, w1.body[0].col_offset) == (2, 4)
        assert (w1.body[0].end_lineno, w1.body[0].end_col_offset) == (2, 8)
        assert (w1.orelse[0].lineno, w1.orelse[0].col_offset) == (4, 4)
        assert (w1.orelse[0].end_lineno, w1.orelse[0].end_col_offset) == (4, 8)

    @staticmethod
    def test_end_lineno_string() -> None:
        """FormattedValue, JoinedStr."""
        code = textwrap.dedent(
            """
        f"Hello World: {42.1234:02d}"  #@
        f"Hello: {name=}"  #@
        """
        ).strip()
        ast_nodes = builder.extract_node(code)
        assert isinstance(ast_nodes, list) and len(ast_nodes) == 2

        s1 = ast_nodes[0]
        assert isinstance(s1, nodes.JoinedStr)
        assert isinstance(s1.values[0], nodes.Const)
        assert (s1.lineno, s1.col_offset) == (1, 0)
        assert (s1.end_lineno, s1.end_col_offset) == (1, 29)
        if PY312_PLUS:
            assert (s1.values[0].lineno, s1.values[0].col_offset) == (1, 2)
            assert (s1.values[0].end_lineno, s1.values[0].end_col_offset) == (1, 15)
        else:
            # Bug in Python 3.11
            # https://github.com/python/cpython/issues/81639
            assert (s1.values[0].lineno, s1.values[0].col_offset) == (1, 0)
            assert (s1.values[0].end_lineno, s1.values[0].end_col_offset) == (1, 29)

        s2 = s1.values[1]
        assert isinstance(s2, nodes.FormattedValue)
        if PY312_PLUS:
            assert (s2.lineno, s2.col_offset) == (1, 15)
            assert (s2.end_lineno, s2.end_col_offset) == (1, 28)
        else:
            assert (s2.lineno, s2.col_offset) == (1, 0)
            assert (s2.end_lineno, s2.end_col_offset) == (1, 29)

        assert isinstance(s2.value, nodes.Const)  # 42.1234
        assert (s2.value.lineno, s2.value.col_offset) == (1, 16)
        assert (s2.value.end_lineno, s2.value.end_col_offset) == (1, 23)
        assert isinstance(s2.format_spec, nodes.JoinedStr)  # ':02d'
        if PY312_PLUS:
            assert (s2.format_spec.lineno, s2.format_spec.col_offset) == (1, 23)
            assert (s2.format_spec.end_lineno, s2.format_spec.end_col_offset) == (1, 27)
        else:
            assert (s2.format_spec.lineno, s2.format_spec.col_offset) == (1, 0)
            assert (s2.format_spec.end_lineno, s2.format_spec.end_col_offset) == (1, 29)

        s3 = ast_nodes[1]
        assert isinstance(s3, nodes.JoinedStr)
        assert isinstance(s3.values[0], nodes.Const)
        assert (s3.lineno, s3.col_offset) == (2, 0)
        assert (s3.end_lineno, s3.end_col_offset) == (2, 17)
        if PY312_PLUS:
            assert (s3.values[0].lineno, s3.values[0].col_offset) == (2, 2)
            assert (s3.values[0].end_lineno, s3.values[0].end_col_offset) == (2, 15)
        else:
            assert (s3.values[0].lineno, s3.values[0].col_offset) == (2, 0)
            assert (s3.values[0].end_lineno, s3.values[0].end_col_offset) == (2, 17)

        s4 = s3.values[1]
        assert isinstance(s4, nodes.FormattedValue)
        if PY312_PLUS:
            assert (s4.lineno, s4.col_offset) == (2, 9)
            assert (s4.end_lineno, s4.end_col_offset) == (2, 16)
        else:
            assert (s4.lineno, s4.col_offset) == (2, 0)
            assert (s4.end_lineno, s4.end_col_offset) == (2, 17)

        assert isinstance(s4.value, nodes.Name)  # 'name'
        assert (s4.value.lineno, s4.value.col_offset) == (2, 10)
        assert (s4.value.end_lineno, s4.value.end_col_offset) == (2, 14)

    @staticmethod
    @pytest.mark.skipif(not PY310_PLUS, reason="pattern matching was added in PY310")
    def test_end_lineno_match() -> None:
        """Match, MatchValue, MatchSingleton, MatchSequence, MatchMapping,
        MatchClass, MatchStar, MatchOr, MatchAs.
        """
        code = textwrap.dedent(
            """
        match x:  #@
            case 200 if True:  #@
                pass
            case True:  #@
                pass
            case (1, 2, *args): #@
                pass
            case {1: "Hello", **rest}: #@
                pass
            case Point2d(0, y=0):  #@
                pass
            case 200 | 300:  #@
                pass
            case 200 as c:  #@
                pass
        """
        ).strip()
        ast_nodes = builder.extract_node(code)
        assert isinstance(ast_nodes, list) and len(ast_nodes) == 8

        # fmt: off
        m1 = ast_nodes[0]
        assert isinstance(m1, nodes.Match)
        assert (m1.lineno, m1.col_offset) == (1, 0)
        assert (m1.end_lineno, m1.end_col_offset) == (15, 12)
        assert (m1.subject.lineno, m1.subject.col_offset) == (1, 6)
        assert (m1.subject.end_lineno, m1.subject.end_col_offset) == (1, 7)

        m2 = ast_nodes[1]
        assert isinstance(m2, nodes.MatchCase)
        assert isinstance(m2.pattern, nodes.MatchValue)
        assert isinstance(m2.guard, nodes.Const)
        assert isinstance(m2.body[0], nodes.Pass)
        assert (m2.pattern.lineno, m2.pattern.col_offset) == (2, 9)
        assert (m2.pattern.end_lineno, m2.pattern.end_col_offset) == (2, 12)
        assert (m2.guard.lineno, m2.guard.col_offset) == (2, 16)
        assert (m2.guard.end_lineno, m2.guard.end_col_offset) == (2, 20)
        assert (m2.body[0].lineno, m2.body[0].col_offset) == (3, 8)
        assert (m2.body[0].end_lineno, m2.body[0].end_col_offset) == (3, 12)

        m3 = ast_nodes[2]
        assert isinstance(m3, nodes.MatchCase)
        assert isinstance(m3.pattern, nodes.MatchSingleton)
        assert (m3.pattern.lineno, m3.pattern.col_offset) == (4, 9)
        assert (m3.pattern.end_lineno, m3.pattern.end_col_offset) == (4, 13)

        m4 = ast_nodes[3]
        assert isinstance(m4, nodes.MatchCase)
        assert isinstance(m4.pattern, nodes.MatchSequence)
        assert isinstance(m4.pattern.patterns[0], nodes.MatchValue)
        assert (m4.pattern.lineno, m4.pattern.col_offset) == (6, 9)
        assert (m4.pattern.end_lineno, m4.pattern.end_col_offset) == (6, 22)
        assert (m4.pattern.patterns[0].lineno, m4.pattern.patterns[0].col_offset) == (6, 10)
        assert (m4.pattern.patterns[0].end_lineno, m4.pattern.patterns[0].end_col_offset) == (6, 11)

        m5 = m4.pattern.patterns[2]
        assert isinstance(m5, nodes.MatchStar)
        assert isinstance(m5.name, nodes.AssignName)
        assert (m5.lineno, m5.col_offset) == (6, 16)
        assert (m5.end_lineno, m5.end_col_offset) == (6, 21)
        # TODO fix column offset: MatchStar -> name (AssignName)
        assert (m5.name.lineno, m5.name.col_offset) == (6, 16)
        assert (m5.name.end_lineno, m5.name.end_col_offset) == (6, 21)

        m6 = ast_nodes[4]
        assert isinstance(m6, nodes.MatchCase)
        assert isinstance(m6.pattern, nodes.MatchMapping)
        assert isinstance(m6.pattern.keys[0], nodes.Const)
        assert isinstance(m6.pattern.patterns[0], nodes.MatchValue)
        assert isinstance(m6.pattern.rest, nodes.AssignName)
        assert (m6.pattern.lineno, m6.pattern.col_offset) == (8, 9)
        assert (m6.pattern.end_lineno, m6.pattern.end_col_offset) == (8, 29)
        assert (m6.pattern.keys[0].lineno, m6.pattern.keys[0].col_offset) == (8, 10)
        assert (m6.pattern.keys[0].end_lineno, m6.pattern.keys[0].end_col_offset) == (8, 11)
        assert (m6.pattern.patterns[0].lineno, m6.pattern.patterns[0].col_offset) == (8, 13)
        assert (m6.pattern.patterns[0].end_lineno, m6.pattern.patterns[0].end_col_offset) == (8, 20)
        # TODO fix column offset: MatchMapping -> rest (AssignName)
        assert (m6.pattern.rest.lineno, m6.pattern.rest.col_offset) == (8, 9)
        assert (m6.pattern.rest.end_lineno, m6.pattern.rest.end_col_offset) == (8, 29)

        m7 = ast_nodes[5]
        assert isinstance(m7, nodes.MatchCase)
        assert isinstance(m7.pattern, nodes.MatchClass)
        assert isinstance(m7.pattern.cls, nodes.Name)
        assert isinstance(m7.pattern.patterns[0], nodes.MatchValue)
        assert isinstance(m7.pattern.kwd_patterns[0], nodes.MatchValue)
        assert (m7.pattern.lineno, m7.pattern.col_offset) == (10, 9)
        assert (m7.pattern.end_lineno, m7.pattern.end_col_offset) == (10, 24)
        assert (m7.pattern.cls.lineno, m7.pattern.cls.col_offset) == (10, 9)
        assert (m7.pattern.cls.end_lineno, m7.pattern.cls.end_col_offset) == (10, 16)
        assert (m7.pattern.patterns[0].lineno, m7.pattern.patterns[0].col_offset) == (10, 17)
        assert (m7.pattern.patterns[0].end_lineno, m7.pattern.patterns[0].end_col_offset) == (10, 18)
        assert (m7.pattern.kwd_patterns[0].lineno, m7.pattern.kwd_patterns[0].col_offset) == (10, 22)
        assert (m7.pattern.kwd_patterns[0].end_lineno, m7.pattern.kwd_patterns[0].end_col_offset) == (10, 23)

        m8 = ast_nodes[6]
        assert isinstance(m8, nodes.MatchCase)
        assert isinstance(m8.pattern, nodes.MatchOr)
        assert isinstance(m8.pattern.patterns[0], nodes.MatchValue)
        assert (m8.pattern.lineno, m8.pattern.col_offset) == (12, 9)
        assert (m8.pattern.end_lineno, m8.pattern.end_col_offset) == (12, 18)
        assert (m8.pattern.patterns[0].lineno, m8.pattern.patterns[0].col_offset) == (12, 9)
        assert (m8.pattern.patterns[0].end_lineno, m8.pattern.patterns[0].end_col_offset) == (12, 12)

        m9 = ast_nodes[7]
        assert isinstance(m9, nodes.MatchCase)
        assert isinstance(m9.pattern, nodes.MatchAs)
        assert isinstance(m9.pattern.pattern, nodes.MatchValue)
        assert isinstance(m9.pattern.name, nodes.AssignName)
        assert (m9.pattern.lineno, m9.pattern.col_offset) == (14, 9)
        assert (m9.pattern.end_lineno, m9.pattern.end_col_offset) == (14, 17)
        assert (m9.pattern.pattern.lineno, m9.pattern.pattern.col_offset) == (14, 9)
        assert (m9.pattern.pattern.end_lineno, m9.pattern.pattern.end_col_offset) == (14, 12)
        # TODO fix column offset: MatchAs -> name (AssignName)
        assert (m9.pattern.name.lineno, m9.pattern.name.col_offset) == (14, 9)
        assert (m9.pattern.name.end_lineno, m9.pattern.name.end_col_offset) == (14, 17)
        # fmt: on

    @staticmethod
    def test_end_lineno_comprehension() -> None:
        """ListComp, SetComp, DictComp, GeneratorExpr."""
        code = textwrap.dedent(
            """
        [x for x in var]  #@
        {x for x in var}  #@
        {x: y for x, y in var}  #@
        (x for x in var)  #@
        """
        ).strip()
        ast_nodes = builder.extract_node(code)
        assert isinstance(ast_nodes, list) and len(ast_nodes) == 4

        c1 = ast_nodes[0]
        assert isinstance(c1, nodes.ListComp)
        assert isinstance(c1.elt, nodes.Name)
        assert isinstance(c1.generators[0], nodes.Comprehension)  # type: ignore[index]
        assert (c1.lineno, c1.col_offset) == (1, 0)
        assert (c1.end_lineno, c1.end_col_offset) == (1, 16)
        assert (c1.elt.lineno, c1.elt.col_offset) == (1, 1)
        assert (c1.elt.end_lineno, c1.elt.end_col_offset) == (1, 2)

        c2 = ast_nodes[1]
        assert isinstance(c2, nodes.SetComp)
        assert isinstance(c2.elt, nodes.Name)
        assert isinstance(c2.generators[0], nodes.Comprehension)  # type: ignore[index]
        assert (c2.lineno, c2.col_offset) == (2, 0)
        assert (c2.end_lineno, c2.end_col_offset) == (2, 16)
        assert (c2.elt.lineno, c2.elt.col_offset) == (2, 1)
        assert (c2.elt.end_lineno, c2.elt.end_col_offset) == (2, 2)

        c3 = ast_nodes[2]
        assert isinstance(c3, nodes.DictComp)
        assert isinstance(c3.key, nodes.Name)
        assert isinstance(c3.value, nodes.Name)
        assert isinstance(c3.generators[0], nodes.Comprehension)  # type: ignore[index]
        assert (c3.lineno, c3.col_offset) == (3, 0)
        assert (c3.end_lineno, c3.end_col_offset) == (3, 22)
        assert (c3.key.lineno, c3.key.col_offset) == (3, 1)
        assert (c3.key.end_lineno, c3.key.end_col_offset) == (3, 2)
        assert (c3.value.lineno, c3.value.col_offset) == (3, 4)
        assert (c3.value.end_lineno, c3.value.end_col_offset) == (3, 5)

        c4 = ast_nodes[3]
        assert isinstance(c4, nodes.GeneratorExp)
        assert isinstance(c4.elt, nodes.Name)
        assert isinstance(c4.generators[0], nodes.Comprehension)  # type: ignore[index]
        assert (c4.lineno, c4.col_offset) == (4, 0)
        assert (c4.end_lineno, c4.end_col_offset) == (4, 16)
        assert (c4.elt.lineno, c4.elt.col_offset) == (4, 1)
        assert (c4.elt.end_lineno, c4.elt.end_col_offset) == (4, 2)

    @staticmethod
    def test_end_lineno_class() -> None:
        """ClassDef, Keyword."""
        code = textwrap.dedent(
            """
        @decorator1
        @decorator2
        class X(Parent, var=42):
            pass
        """
        ).strip()
        c1 = builder.extract_node(code)
        assert isinstance(c1, nodes.ClassDef)
        assert isinstance(c1.decorators, nodes.Decorators)
        assert isinstance(c1.bases[0], nodes.Name)
        assert isinstance(c1.keywords[0], nodes.Keyword)
        assert isinstance(c1.body[0], nodes.Pass)

        # fmt: off
        assert (c1.lineno, c1.col_offset) == (3, 0)
        assert (c1.end_lineno, c1.end_col_offset) == (4, 8)
        assert (c1.decorators.lineno, c1.decorators.col_offset) == (1, 0)
        assert (c1.decorators.end_lineno, c1.decorators.end_col_offset) == (2, 11)
        assert (c1.bases[0].lineno, c1.bases[0].col_offset) == (3, 8)
        assert (c1.bases[0].end_lineno, c1.bases[0].end_col_offset) == (3, 14)
        assert (c1.keywords[0].lineno, c1.keywords[0].col_offset) == (3, 16)
        assert (c1.keywords[0].end_lineno, c1.keywords[0].end_col_offset) == (3, 22)
        assert (c1.body[0].lineno, c1.body[0].col_offset) == (4, 4)
        assert (c1.body[0].end_lineno, c1.body[0].end_col_offset) == (4, 8)
        # fmt: on

    @staticmethod
    def test_end_lineno_module() -> None:
        """Tests for Module."""
        code = """print()"""
        module = astroid.parse(code)
        assert isinstance(module, nodes.Module)
        assert module.lineno == 0
        assert module.col_offset == 0
        assert module.end_lineno is None
        assert module.end_col_offset is None



# Licensed under the LGPL: https://www.gnu.org/licenses/old-licenses/lgpl-2.1.en.html
# For details: https://github.com/pylint-dev/astroid/blob/main/LICENSE
# Copyright (c) https://github.com/pylint-dev/astroid/blob/main/CONTRIBUTORS.txt

from __future__ import annotations

import unittest

import pytest

import astroid
from astroid import bases, builder, nodes, objects
from astroid.exceptions import InferenceError


class EnumBrainTest(unittest.TestCase):
    def test_simple_enum(self) -> None:
        module = builder.parse(
            """
        import enum

        class MyEnum(enum.Enum):
            one = "one"
            two = "two"

            def mymethod(self, x):
                return 5

        """
        )

        enumeration = next(module["MyEnum"].infer())
        one = enumeration["one"]
        self.assertEqual(one.pytype(), ".MyEnum.one")

        for propname in ("name", "value"):
            prop = next(iter(one.getattr(propname)))
            self.assertIn("builtins.property", prop.decoratornames())

        meth = one.getattr("mymethod")[0]
        self.assertIsInstance(meth, astroid.FunctionDef)

    def test_looks_like_enum_false_positive(self) -> None:
        # Test that a class named Enumeration is not considered a builtin enum.
        module = builder.parse(
            """
        class Enumeration(object):
            def __init__(self, name, enum_list):
                pass
            test = 42
        """
        )
        enumeration = module["Enumeration"]
        test = next(enumeration.igetattr("test"))
        self.assertEqual(test.value, 42)

    def test_user_enum_false_positive(self) -> None:
        # Test that a user-defined class named Enum is not considered a builtin enum.
        ast_node = astroid.extract_node(
            """
        class Enum:
            pass

        class Color(Enum):
            red = 1

        Color.red #@
        """
        )
        assert isinstance(ast_node, nodes.NodeNG)
        inferred = ast_node.inferred()
        self.assertEqual(len(inferred), 1)
        self.assertIsInstance(inferred[0], astroid.Const)
        self.assertEqual(inferred[0].value, 1)

    def test_ignores_with_nodes_from_body_of_enum(self) -> None:
        code = """
        import enum

        class Error(enum.Enum):
            Foo = "foo"
            Bar = "bar"
            with "error" as err:
                pass
        """
        node = builder.extract_node(code)
        inferred = next(node.infer())
        assert "err" in inferred.locals
        assert len(inferred.locals["err"]) == 1

    def test_enum_multiple_base_classes(self) -> None:
        module = builder.parse(
            """
        import enum

        class Mixin:
            pass

        class MyEnum(Mixin, enum.Enum):
            one = 1
        """
        )
        enumeration = next(module["MyEnum"].infer())
        one = enumeration["one"]

        clazz = one.getattr("__class__")[0]
        self.assertTrue(
            clazz.is_subtype_of(".Mixin"),
            "Enum instance should share base classes with generating class",
        )

    def test_int_enum(self) -> None:
        module = builder.parse(
            """
        import enum

        class MyEnum(enum.IntEnum):
            one = 1
        """
        )

        enumeration = next(module["MyEnum"].infer())
        one = enumeration["one"]

        clazz = one.getattr("__class__")[0]
        self.assertTrue(
            clazz.is_subtype_of("builtins.int"),
            "IntEnum based enums should be a subtype of int",
        )

    def test_enum_func_form_is_class_not_instance(self) -> None:
        cls, instance = builder.extract_node(
            """
        from enum import Enum
        f = Enum('Audience', ['a', 'b', 'c'])
        f #@
        f(1) #@
        """
        )
        inferred_cls = next(cls.infer())
        self.assertIsInstance(inferred_cls, bases.Instance)
        inferred_instance = next(instance.infer())
        self.assertIsInstance(inferred_instance, bases.Instance)
        self.assertIsInstance(next(inferred_instance.igetattr("name")), nodes.Const)
        self.assertIsInstance(next(inferred_instance.igetattr("value")), nodes.Const)

    def test_enum_func_form_iterable(self) -> None:
        instance = builder.extract_node(
            """
        from enum import Enum
        Animal = Enum('Animal', 'ant bee cat dog')
        Animal
        """
        )
        inferred = next(instance.infer())
        self.assertIsInstance(inferred, astroid.Instance)
        self.assertTrue(inferred.getattr("__iter__"))

    def test_enum_func_form_subscriptable(self) -> None:
        instance, name = builder.extract_node(
            """
        from enum import Enum
        Animal = Enum('Animal', 'ant bee cat dog')
        Animal['ant'] #@
        Animal['ant'].name #@
        """
        )
        instance = next(instance.infer())
        self.assertIsInstance(instance, astroid.Instance)

        inferred = next(name.infer())
        self.assertIsInstance(inferred, astroid.Const)

    def test_enum_func_form_has_dunder_members(self) -> None:
        instance = builder.extract_node(
            """
        from enum import Enum
        Animal = Enum('Animal', 'ant bee cat dog')
        for i in Animal.__members__:
            i #@
        """
        )
        instance = next(instance.infer())
        self.assertIsInstance(instance, astroid.Const)
        self.assertIsInstance(instance.value, str)

    def test_infer_enum_value_as_the_right_type(self) -> None:
        string_value, int_value = builder.extract_node(
            """
        from enum import Enum
        class A(Enum):
            a = 'a'
            b = 1
        A.a.value #@
        A.b.value #@
        """
        )
        inferred_string = string_value.inferred()
        assert any(
            isinstance(elem, astroid.Const) and elem.value == "a"
            for elem in inferred_string
        )

        inferred_int = int_value.inferred()
        assert any(
            isinstance(elem, astroid.Const) and elem.value == 1 for elem in inferred_int
        )

    def test_mingled_single_and_double_quotes_does_not_crash(self) -> None:
        node = builder.extract_node(
            """
        from enum import Enum
        class A(Enum):
            a = 'x"y"'
        A.a.value #@
        """
        )
        inferred_string = next(node.infer())
        assert inferred_string.value == 'x"y"'

    def test_special_characters_does_not_crash(self) -> None:
        node = builder.extract_node(
            """
        import enum
        class Example(enum.Enum):
            NULL = '\\N{NULL}'
        Example.NULL.value
        """
        )
        inferred_string = next(node.infer())
        assert inferred_string.value == "\N{NULL}"

    def test_dont_crash_on_for_loops_in_body(self) -> None:
        node = builder.extract_node(
            """

        class Commands(IntEnum):
            _ignore_ = 'Commands index'
            _init_ = 'value string'

            BEL = 0x07, 'Bell'
            Commands = vars()
            for index in range(4):
                Commands[f'DC{index + 1}'] = 0x11 + index, f'Device Control {index + 1}'

        Commands
        """
        )
        inferred = next(node.infer())
        assert isinstance(inferred, astroid.ClassDef)

    def test_enum_tuple_list_values(self) -> None:
        tuple_node, list_node = builder.extract_node(
            """
        import enum

        class MyEnum(enum.Enum):
            a = (1, 2)
            b = [2, 4]
        MyEnum.a.value #@
        MyEnum.b.value #@
        """
        )
        inferred_tuple_node = next(tuple_node.infer())
        inferred_list_node = next(list_node.infer())
        assert isinstance(inferred_tuple_node, astroid.Tuple)
        assert isinstance(inferred_list_node, astroid.List)
        assert inferred_tuple_node.as_string() == "(1, 2)"
        assert inferred_list_node.as_string() == "[2, 4]"

    def test_enum_starred_is_skipped(self) -> None:
        code = """
        from enum import Enum
        class ContentType(Enum):
            TEXT, PHOTO, VIDEO, GIF, YOUTUBE, *_ = [1, 2, 3, 4, 5, 6]
        ContentType.TEXT #@
        """
        node = astroid.extract_node(code)
        next(node.infer())

    def test_enum_name_is_str_on_self(self) -> None:
        code = """
        from enum import Enum
        class TestEnum(Enum):
            def func(self):
                self.name #@
                self.value #@
        TestEnum.name #@
        TestEnum.value #@
        """
        i_name, i_value, c_name, c_value = astroid.extract_node(code)

        # <instance>.name should be a string, <class>.name should be a property (that
        # forwards the lookup to __getattr__)
        inferred = next(i_name.infer())
        assert isinstance(inferred, nodes.Const)
        assert inferred.pytype() == "builtins.str"
        inferred = next(c_name.infer())
        assert isinstance(inferred, objects.Property)

        # Inferring .value should not raise InferenceError. It is probably Uninferable
        # but we don't particularly care
        next(i_value.infer())
        next(c_value.infer())

    def test_enum_name_and_value_members_override_dynamicclassattr(self) -> None:
        code = """
        from enum import Enum
        class TrickyEnum(Enum):
            name = 1
            value = 2

            def func(self):
                self.name #@
                self.value #@
        TrickyEnum.name #@
        TrickyEnum.value #@
        """
        i_name, i_value, c_name, c_value = astroid.extract_node(code)

        # All of these cases should be inferred as enum members
        inferred = next(i_name.infer())
        assert isinstance(inferred, bases.Instance)
        assert inferred.pytype() == ".TrickyEnum.name"
        inferred = next(c_name.infer())
        assert isinstance(inferred, bases.Instance)
        assert inferred.pytype() == ".TrickyEnum.name"
        inferred = next(i_value.infer())
        assert isinstance(inferred, bases.Instance)
        assert inferred.pytype() == ".TrickyEnum.value"
        inferred = next(c_value.infer())
        assert isinstance(inferred, bases.Instance)
        assert inferred.pytype() == ".TrickyEnum.value"

    def test_enum_subclass_member_name(self) -> None:
        ast_node = astroid.extract_node(
            """
        from enum import Enum

        class EnumSubclass(Enum):
            pass

        class Color(EnumSubclass):
            red = 1

        Color.red.name #@
        """
        )
        assert isinstance(ast_node, nodes.NodeNG)
        inferred = ast_node.inferred()
        self.assertEqual(len(inferred), 1)
        self.assertIsInstance(inferred[0], astroid.Const)
        self.assertEqual(inferred[0].value, "red")

    def test_enum_subclass_member_value(self) -> None:
        ast_node = astroid.extract_node(
            """
        from enum import Enum

        class EnumSubclass(Enum):
            pass

        class Color(EnumSubclass):
            red = 1

        Color.red.value #@
        """
        )
        assert isinstance(ast_node, nodes.NodeNG)
        inferred = ast_node.inferred()
        self.assertEqual(len(inferred), 1)
        self.assertIsInstance(inferred[0], astroid.Const)
        self.assertEqual(inferred[0].value, 1)

    def test_enum_subclass_member_method(self) -> None:
        # See Pylint issue #2626
        ast_node = astroid.extract_node(
            """
        from enum import Enum

        class EnumSubclass(Enum):
            def hello_pylint(self) -> str:
                return self.name

        class Color(EnumSubclass):
            red = 1

        Color.red.hello_pylint()  #@
        """
        )
        assert isinstance(ast_node, nodes.NodeNG)
        inferred = ast_node.inferred()
        self.assertEqual(len(inferred), 1)
        self.assertIsInstance(inferred[0], astroid.Const)
        self.assertEqual(inferred[0].value, "red")

    def test_enum_subclass_different_modules(self) -> None:
        # See Pylint issue #2626
        astroid.extract_node(
            """
        from enum import Enum

        class EnumSubclass(Enum):
            pass
        """,
            "a",
        )
        ast_node = astroid.extract_node(
            """
        from a import EnumSubclass

        class Color(EnumSubclass):
            red = 1

        Color.red.value #@
        """
        )
        assert isinstance(ast_node, nodes.NodeNG)
        inferred = ast_node.inferred()
        self.assertEqual(len(inferred), 1)
        self.assertIsInstance(inferred[0], astroid.Const)
        self.assertEqual(inferred[0].value, 1)

    def test_members_member_ignored(self) -> None:
        ast_node = builder.extract_node(
            """
        from enum import Enum
        class Animal(Enum):
            a = 1
            __members__ = {}
        Animal.__members__ #@
        """
        )

        inferred = next(ast_node.infer())
        self.assertIsInstance(inferred, astroid.Dict)
        self.assertTrue(inferred.locals)

    def test_enum_as_renamed_import(self) -> None:
        """Originally reported in https://github.com/pylint-dev/pylint/issues/5776."""
        ast_node: nodes.Attribute = builder.extract_node(
            """
        from enum import Enum as PyEnum
        class MyEnum(PyEnum):
            ENUM_KEY = "enum_value"
        MyEnum.ENUM_KEY
        """
        )
        inferred = next(ast_node.infer())
        assert isinstance(inferred, bases.Instance)
        assert inferred._proxied.name == "ENUM_KEY"

    def test_class_named_enum(self) -> None:
        """Test that the user-defined class named `Enum` is not inferred as `enum.Enum`"""
        astroid.extract_node(
            """
        class Enum:
            def __init__(self, one, two):
                self.one = one
                self.two = two
            def pear(self):
                ...
        """,
            "module_with_class_named_enum",
        )

        attribute_nodes = astroid.extract_node(
            """
        import module_with_class_named_enum
        module_with_class_named_enum.Enum("apple", "orange") #@
        typo_module_with_class_named_enum.Enum("apple", "orange") #@
        """
        )

        name_nodes = astroid.extract_node(
            """
        from module_with_class_named_enum import Enum
        Enum("apple", "orange") #@
        TypoEnum("apple", "orange") #@
        """
        )

        # Test that both of the successfully inferred `Name` & `Attribute`
        # nodes refer to the user-defined Enum class.
        for inferred in (attribute_nodes[0].inferred()[0], name_nodes[0].inferred()[0]):
            assert isinstance(inferred, astroid.Instance)
            assert inferred.name == "Enum"
            assert inferred.qname() == "module_with_class_named_enum.Enum"
            assert "pear" in inferred.locals

        # Test that an `InferenceError` is raised when an attempt is made to
        # infer a `Name` or `Attribute` node & they cannot be found.
        for node in (attribute_nodes[1], name_nodes[1]):
            with pytest.raises(InferenceError):
                node.inferred()

    def test_enum_members_uppercase_only(self) -> None:
        """Originally reported in https://github.com/pylint-dev/pylint/issues/7402.
        ``nodes.AnnAssign`` nodes with no assigned values do not appear inside ``__members__``.

        Test that only enum members `MARS` and `radius` appear in the `__members__` container while
        the attribute `mass` does not.
        """
        enum_class = astroid.extract_node(
            """
        from enum import Enum
        class Planet(Enum): #@
            MARS = (1, 2)
            radius: int = 1
            mass: int

            def __init__(self, mass, radius):
                self.mass = mass
                self.radius = radius

        Planet.MARS.value
        """
        )
        enum_members = next(enum_class.igetattr("__members__"))
        assert len(enum_members.items) == 2
        mars, radius = enum_members.items
        assert mars[1].name == "MARS"
        assert radius[1].name == "radius"

    def test_local_enum_child_class_inference(self) -> None:
        """Originally reported in https://github.com/pylint-dev/pylint/issues/8897

        Test that a user-defined enum class is inferred when it subclasses
        another user-defined enum class.
        """
        enum_class_node, enum_member_value_node = astroid.extract_node(
            """
        import sys

        from enum import Enum

        if sys.version_info >= (3, 11):
            from enum import StrEnum
        else:
            class StrEnum(str, Enum):
                pass


        class Color(StrEnum): #@
            RED = "red"


        Color.RED.value #@
        """
        )
        assert "RED" in enum_class_node.locals

        enum_members = enum_class_node.locals["__members__"][0].items
        assert len(enum_members) == 1
        _, name = enum_members[0]
        assert name.name == "RED"

        inferred_enum_member_value_node = next(enum_member_value_node.infer())
        assert inferred_enum_member_value_node.value == "red"

    def test_enum_with_ignore(self) -> None:
        """Exclude ``_ignore_`` from the ``__members__`` container
        Originally reported in https://github.com/pylint-dev/pylint/issues/9015
        """

        ast_node: nodes.Attribute = builder.extract_node(
            """
        import enum


        class MyEnum(enum.Enum):
            FOO = enum.auto()
            BAR = enum.auto()
            _ignore_ = ["BAZ"]
            BAZ = 42
        MyEnum.__members__
        """
        )
        inferred = next(ast_node.infer())
        members_names = [const_node.value for const_node, name_obj in inferred.items]
        assert members_names == ["FOO", "BAR", "BAZ"]

    def test_enum_sunder_names(self) -> None:
        """Test that both `_name_` and `_value_` sunder names exist"""

        sunder_name, sunder_value = builder.extract_node(
            """
        import enum


        class MyEnum(enum.Enum):
            APPLE = 42
        MyEnum.APPLE._name_ #@
        MyEnum.APPLE._value_ #@
        """
        )
        inferred_name = next(sunder_name.infer())
        assert inferred_name.value == "APPLE"

        inferred_value = next(sunder_value.infer())
        assert inferred_value.value == 42



# Licensed under the LGPL: https://www.gnu.org/licenses/old-licenses/lgpl-2.1.en.html
# For details: https://github.com/pylint-dev/astroid/blob/main/LICENSE
# Copyright (c) https://github.com/pylint-dev/astroid/blob/main/CONTRIBUTORS.txt

import os
import site
import sys
import time
import unittest
import warnings
from collections.abc import Iterator
from contextlib import contextmanager
from unittest import mock

import pytest

import astroid
from astroid import manager, test_utils
from astroid.const import IS_JYTHON, IS_PYPY, PY312_PLUS
from astroid.exceptions import (
    AstroidBuildingError,
    AstroidImportError,
    AttributeInferenceError,
)
from astroid.interpreter._import import util
from astroid.modutils import EXT_LIB_DIRS, module_in_path
from astroid.nodes import Const
from astroid.nodes.scoped_nodes import ClassDef, Module

from . import resources


def _get_file_from_object(obj) -> str:
    if IS_JYTHON:
        return obj.__file__.split("$py.class")[0] + ".py"
    return obj.__file__


class AstroidManagerTest(resources.SysPathSetup, unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.manager = test_utils.brainless_manager()
        self.manager.clear_cache()

    def test_ast_from_file(self) -> None:
        filepath = unittest.__file__
        ast = self.manager.ast_from_file(filepath)
        self.assertEqual(ast.name, "unittest")
        self.assertIn("unittest", self.manager.astroid_cache)

    def test_ast_from_file_cache(self) -> None:
        filepath = unittest.__file__
        self.manager.ast_from_file(filepath)
        ast = self.manager.ast_from_file("unhandledName", "unittest")
        self.assertEqual(ast.name, "unittest")
        self.assertIn("unittest", self.manager.astroid_cache)

    def test_ast_from_file_astro_builder(self) -> None:
        filepath = unittest.__file__
        ast = self.manager.ast_from_file(filepath, None, True, True)
        self.assertEqual(ast.name, "unittest")
        self.assertIn("unittest", self.manager.astroid_cache)

    def test_ast_from_file_name_astro_builder_exception(self) -> None:
        self.assertRaises(
            AstroidBuildingError, self.manager.ast_from_file, "unhandledName"
        )

    def test_ast_from_string(self) -> None:
        filepath = unittest.__file__
        dirname = os.path.dirname(filepath)
        modname = os.path.basename(dirname)
        with open(filepath, encoding="utf-8") as file:
            data = file.read()
            ast = self.manager.ast_from_string(data, modname, filepath)
            self.assertEqual(ast.name, "unittest")
            self.assertEqual(ast.file, filepath)
            self.assertIn("unittest", self.manager.astroid_cache)

    def test_do_not_expose_main(self) -> None:
        obj = self.manager.ast_from_module_name("__main__")
        self.assertEqual(obj.name, "__main__")
        self.assertEqual(obj.items(), [])

    def test_ast_from_module_name(self) -> None:
        ast = self.manager.ast_from_module_name("unittest")
        self.assertEqual(ast.name, "unittest")
        self.assertIn("unittest", self.manager.astroid_cache)

    def test_ast_from_module_name_not_python_source(self) -> None:
        ast = self.manager.ast_from_module_name("time")
        self.assertEqual(ast.name, "time")
        self.assertIn("time", self.manager.astroid_cache)
        self.assertEqual(ast.pure_python, False)

    def test_ast_from_module_name_astro_builder_exception(self) -> None:
        self.assertRaises(
            AstroidBuildingError,
            self.manager.ast_from_module_name,
            "unhandledModule",
        )

    def _test_ast_from_old_namespace_package_protocol(self, root: str) -> None:
        origpath = sys.path[:]
        paths = [resources.find(f"data/path_{root}_{index}") for index in range(1, 4)]
        sys.path.extend(paths)
        try:
            for name in ("foo", "bar", "baz"):
                module = self.manager.ast_from_module_name("package." + name)
                self.assertIsInstance(module, astroid.Module)
        finally:
            sys.path = origpath

    def test_ast_from_namespace_pkgutil(self) -> None:
        self._test_ast_from_old_namespace_package_protocol("pkgutil")

    def test_ast_from_namespace_pkg_resources(self) -> None:
        self._test_ast_from_old_namespace_package_protocol("pkg_resources")

    def test_identify_old_namespace_package_protocol(self) -> None:
        # Like the above cases, this package follows the old namespace package protocol
        # astroid currently assumes such packages are in sys.modules, so import it
        # pylint: disable-next=import-outside-toplevel
        import tests.testdata.python3.data.path_pkg_resources_1.package.foo as _  # noqa

        self.assertTrue(
            util.is_namespace("tests.testdata.python3.data.path_pkg_resources_1")
        )

    def test_submodule_homonym_with_non_module(self) -> None:
        self.assertFalse(
            util.is_namespace("tests.testdata.python3.data.parent_of_homonym.doc")
        )

    def test_module_is_not_namespace(self) -> None:
        self.assertFalse(util.is_namespace("tests.testdata.python3.data.all"))
        self.assertFalse(util.is_namespace("__main__"))
        self.assertFalse(
            util.is_namespace(next(iter(EXT_LIB_DIRS)).rsplit("/", maxsplit=1)[-1]),
        )
        self.assertFalse(util.is_namespace("importlib._bootstrap"))

    def test_module_unexpectedly_missing_spec(self) -> None:
        astroid_module = sys.modules["astroid"]
        original_spec = astroid_module.__spec__
        del astroid_module.__spec__
        try:
            self.assertFalse(util.is_namespace("astroid"))
        finally:
            astroid_module.__spec__ = original_spec

    @mock.patch(
        "astroid.interpreter._import.util._find_spec_from_path",
        side_effect=AttributeError,
    )
    def test_module_unexpectedly_missing_path(self, mocked) -> None:
        """Https://github.com/pylint-dev/pylint/issues/7592."""
        self.assertFalse(util.is_namespace("astroid"))

    def test_module_unexpectedly_spec_is_none(self) -> None:
        astroid_module = sys.modules["astroid"]
        original_spec = astroid_module.__spec__
        astroid_module.__spec__ = None
        try:
            self.assertFalse(util.is_namespace("astroid"))
        finally:
            astroid_module.__spec__ = original_spec

    def test_implicit_namespace_package(self) -> None:
        data_dir = os.path.dirname(resources.find("data/namespace_pep_420"))
        contribute = os.path.join(data_dir, "contribute_to_namespace")
        for value in (data_dir, contribute):
            sys.path.insert(0, value)

        try:
            module = self.manager.ast_from_module_name("namespace_pep_420.module")
            self.assertIsInstance(module, astroid.Module)
            self.assertEqual(module.name, "namespace_pep_420.module")
            var = next(module.igetattr("var"))
            self.assertIsInstance(var, astroid.Const)
            self.assertEqual(var.value, 42)
        finally:
            for _ in range(2):
                sys.path.pop(0)

    @pytest.mark.skipif(
        IS_PYPY,
        reason="PyPy provides no way to tell apart frozen stdlib from old-style namespace packages",
    )
    def test_namespace_package_pth_support(self) -> None:
        pth = "foogle_fax-0.12.5-py2.7-nspkg.pth"
        site.addpackage(resources.RESOURCE_PATH, pth, [])

        try:
            module = self.manager.ast_from_module_name("foogle.fax")
            submodule = next(module.igetattr("a"))
            value = next(submodule.igetattr("x"))
            self.assertIsInstance(value, astroid.Const)
            with self.assertRaises(AstroidImportError):
                self.manager.ast_from_module_name("foogle.moogle")
        finally:
            sys.modules.pop("foogle")

    @pytest.mark.skipif(
        IS_PYPY,
        reason="PyPy provides no way to tell apart frozen stdlib from old-style namespace packages",
    )
    def test_nested_namespace_import(self) -> None:
        pth = "foogle_fax-0.12.5-py2.7-nspkg.pth"
        site.addpackage(resources.RESOURCE_PATH, pth, [])
        try:
            self.manager.ast_from_module_name("foogle.crank")
        finally:
            sys.modules.pop("foogle")

    def test_namespace_and_file_mismatch(self) -> None:
        filepath = unittest.__file__
        ast = self.manager.ast_from_file(filepath)
        self.assertEqual(ast.name, "unittest")
        pth = "foogle_fax-0.12.5-py2.7-nspkg.pth"
        site.addpackage(resources.RESOURCE_PATH, pth, [])
        try:
            with self.assertRaises(AstroidImportError):
                self.manager.ast_from_module_name("unittest.foogle.fax")
        finally:
            sys.modules.pop("foogle")

    def _test_ast_from_zip(self, archive: str) -> None:
        sys.modules.pop("mypypa", None)
        archive_path = resources.find(archive)
        sys.path.insert(0, archive_path)
        module = self.manager.ast_from_module_name("mypypa")
        self.assertEqual(module.name, "mypypa")
        end = os.path.join(archive, "mypypa")
        self.assertTrue(
            module.file.endswith(end), f"{module.file} doesn't endswith {end}"
        )

    @contextmanager
    def _restore_package_cache(self) -> Iterator:
        orig_path = sys.path[:]
        orig_pathcache = sys.path_importer_cache.copy()
        orig_modcache = self.manager.astroid_cache.copy()
        orig_modfilecache = self.manager._mod_file_cache.copy()
        orig_importhooks = self.manager._failed_import_hooks[:]
        yield
        self.manager._failed_import_hooks = orig_importhooks
        self.manager._mod_file_cache = orig_modfilecache
        self.manager.astroid_cache = orig_modcache
        sys.path_importer_cache = orig_pathcache
        sys.path = orig_path

    def test_ast_from_module_name_egg(self) -> None:
        with self._restore_package_cache():
            self._test_ast_from_zip(
                os.path.sep.join(["data", os.path.normcase("MyPyPa-0.1.0-py2.5.egg")])
            )

    def test_ast_from_module_name_zip(self) -> None:
        with self._restore_package_cache():
            self._test_ast_from_zip(
                os.path.sep.join(["data", os.path.normcase("MyPyPa-0.1.0-py2.5.zip")])
            )

    def test_ast_from_module_name_pyz(self) -> None:
        try:
            linked_file_name = os.path.join(
                resources.RESOURCE_PATH, "MyPyPa-0.1.0-py2.5.pyz"
            )
            os.symlink(
                os.path.join(resources.RESOURCE_PATH, "MyPyPa-0.1.0-py2.5.zip"),
                linked_file_name,
            )

            with self._restore_package_cache():
                self._test_ast_from_zip(linked_file_name)
        finally:
            os.remove(linked_file_name)

    def test_zip_import_data(self) -> None:
        """Check if zip_import_data works."""
        with self._restore_package_cache():
            filepath = resources.find("data/MyPyPa-0.1.0-py2.5.zip/mypypa")
            ast = self.manager.zip_import_data(filepath)
            self.assertEqual(ast.name, "mypypa")

    def test_zip_import_data_without_zipimport(self) -> None:
        """Check if zip_import_data return None without zipimport."""
        self.assertEqual(self.manager.zip_import_data("path"), None)

    def test_file_from_module(self) -> None:
        """Check if the unittest filepath is equals to the result of the method."""
        self.assertEqual(
            _get_file_from_object(unittest),
            self.manager.file_from_module_name("unittest", None).location,
        )

    def test_file_from_module_name_astro_building_exception(self) -> None:
        """Check if the method raises an exception with a wrong module name."""
        self.assertRaises(
            AstroidBuildingError,
            self.manager.file_from_module_name,
            "unhandledModule",
            None,
        )

    def test_ast_from_module(self) -> None:
        ast = self.manager.ast_from_module(unittest)
        self.assertEqual(ast.pure_python, True)
        ast = self.manager.ast_from_module(time)
        self.assertEqual(ast.pure_python, False)

    def test_ast_from_module_cache(self) -> None:
        """Check if the module is in the cache manager."""
        ast = self.manager.ast_from_module(unittest)
        self.assertEqual(ast.name, "unittest")
        self.assertIn("unittest", self.manager.astroid_cache)

    def test_ast_from_class(self) -> None:
        ast = self.manager.ast_from_class(int)
        self.assertEqual(ast.name, "int")
        self.assertEqual(ast.parent.frame().name, "builtins")
        self.assertEqual(ast.parent.frame().name, "builtins")

        ast = self.manager.ast_from_class(object)
        self.assertEqual(ast.name, "object")
        self.assertEqual(ast.parent.frame().name, "builtins")
        self.assertEqual(ast.parent.frame().name, "builtins")
        self.assertIn("__setattr__", ast)

    def test_ast_from_class_with_module(self) -> None:
        """Check if the method works with the module name."""
        ast = self.manager.ast_from_class(int, int.__module__)
        self.assertEqual(ast.name, "int")
        self.assertEqual(ast.parent.frame().name, "builtins")
        self.assertEqual(ast.parent.frame().name, "builtins")

        ast = self.manager.ast_from_class(object, object.__module__)
        self.assertEqual(ast.name, "object")
        self.assertEqual(ast.parent.frame().name, "builtins")
        self.assertEqual(ast.parent.frame().name, "builtins")
        self.assertIn("__setattr__", ast)

    def test_ast_from_class_attr_error(self) -> None:
        """Give a wrong class at the ast_from_class method."""
        self.assertRaises(AstroidBuildingError, self.manager.ast_from_class, None)

    def test_failed_import_hooks(self) -> None:
        def hook(modname: str):
            if modname == "foo.bar":
                return unittest

            raise AstroidBuildingError()

        with self.assertRaises(AstroidBuildingError):
            self.manager.ast_from_module_name("foo.bar")

        with self._restore_package_cache():
            self.manager.register_failed_import_hook(hook)
            self.assertEqual(unittest, self.manager.ast_from_module_name("foo.bar"))
            with self.assertRaises(AstroidBuildingError):
                self.manager.ast_from_module_name("foo.bar.baz")

    def test_same_name_import_module(self) -> None:
        """Test inference of an import statement with the same name as the module.

        See https://github.com/pylint-dev/pylint/issues/5151.
        """
        math_file = resources.find("data/import_conflicting_names/math.py")
        module = self.manager.ast_from_file(math_file)

        # Change the cache key and module name to mimic importing the test file
        # from the root/top level. This creates a clash between math.py and stdlib math.
        self.manager.astroid_cache["math"] = self.manager.astroid_cache.pop(module.name)
        module.name = "math"

        # Infer the 'import math' statement
        stdlib_math = next(module.body[1].value.args[0].infer())
        assert self.manager.astroid_cache["math"] != stdlib_math

    def test_raises_exception_for_empty_modname(self) -> None:
        with pytest.raises(AstroidBuildingError):
            self.manager.ast_from_module_name(None)

    def test_denied_modules_raise(self) -> None:
        self.manager.module_denylist.add("random")
        with pytest.raises(AstroidImportError, match="random"):
            self.manager.ast_from_module_name("random")
        # and module not in the deny list shouldn't raise
        self.manager.ast_from_module_name("math")


class IsolatedAstroidManagerTest(unittest.TestCase):
    @pytest.mark.skipif(PY312_PLUS, reason="distutils was removed in python 3.12")
    def test_no_user_warning(self):
        """When Python 3.12 is minimum, this test will no longer provide value."""
        mgr = manager.AstroidManager()
        self.addCleanup(mgr.clear_cache)
        with warnings.catch_warnings():
            warnings.filterwarnings("error", category=UserWarning)
            mgr.ast_from_module_name("setuptools")
            try:
                mgr.ast_from_module_name("pip")
            except astroid.AstroidImportError:
                pytest.skip("pip is not installed")


class BorgAstroidManagerTC(unittest.TestCase):
    def test_borg(self) -> None:
        """Test that the AstroidManager is really a borg, i.e. that two different
        instances has same cache.
        """
        first_manager = manager.AstroidManager()
        built = first_manager.ast_from_module_name("builtins")

        second_manager = manager.AstroidManager()
        second_built = second_manager.ast_from_module_name("builtins")
        self.assertIs(built, second_built)

    def test_max_inferable_values(self) -> None:
        mgr = manager.AstroidManager()
        original_limit = mgr.max_inferable_values

        def reset_limit():
            nonlocal original_limit
            manager.AstroidManager().max_inferable_values = original_limit

        self.addCleanup(reset_limit)

        mgr.max_inferable_values = 4
        self.assertEqual(manager.AstroidManager.brain["max_inferable_values"], 4)


class ClearCacheTest(unittest.TestCase):
    def test_clear_cache_clears_other_lru_caches(self) -> None:
        lrus = (
            astroid.nodes._base_nodes.LookupMixIn.lookup,
            astroid.modutils._cache_normalize_path_,
            util.is_namespace,
            astroid.interpreter.objectmodel.ObjectModel.attributes,
            ClassDef._metaclass_lookup_attribute,
        )

        # Get a baseline for the size of the cache after simply calling bootstrap()
        baseline_cache_infos = [lru.cache_info() for lru in lrus]

        # Generate some hits and misses
        module = Module("", file="", path=[], package=False)
        ClassDef(
            "",
            lineno=0,
            col_offset=0,
            end_lineno=0,
            end_col_offset=0,
            parent=module,
        ).lookup("garbage")
        module_in_path("unittest", "garbage_path")
        util.is_namespace("unittest")
        astroid.interpreter.objectmodel.ObjectModel().attributes()
        with pytest.raises(AttributeInferenceError):
            ClassDef(
                "",
                lineno=0,
                col_offset=0,
                end_lineno=0,
                end_col_offset=0,
                parent=module,
            ).getattr("garbage")

        # Did the hits or misses actually happen?
        incremented_cache_infos = [lru.cache_info() for lru in lrus]
        for incremented_cache, baseline_cache in zip(
            incremented_cache_infos, baseline_cache_infos
        ):
            with self.subTest(incremented_cache=incremented_cache):
                self.assertGreater(
                    incremented_cache.hits + incremented_cache.misses,
                    baseline_cache.hits + baseline_cache.misses,
                )

        astroid.MANAGER.clear_cache()  # also calls bootstrap()

        self.assertEqual(astroid.context._INFERENCE_CACHE, {})

        # The cache sizes are now as low or lower than the original baseline
        cleared_cache_infos = [lru.cache_info() for lru in lrus]
        for cleared_cache, baseline_cache in zip(
            cleared_cache_infos, baseline_cache_infos
        ):
            with self.subTest(cleared_cache=cleared_cache):
                # less equal because the "baseline" might have had multiple calls to bootstrap()
                self.assertLessEqual(cleared_cache.currsize, baseline_cache.currsize)

    def test_file_cache_after_clear_cache(self) -> None:
        """Test to mimic the behavior of how pylint lints file and
        ensure clear cache clears everything stored in the cache.
        See https://github.com/pylint-dev/pylint/pull/9932#issuecomment-2364985551
        for more information.
        """
        orig_sys_path = sys.path[:]
        try:
            search_path = resources.RESOURCE_PATH
            sys.path.insert(0, search_path)
            node = astroid.MANAGER.ast_from_file(resources.find("data/cache/a.py"))
            self.assertEqual(node.name, "cache.a")

            # This import from statement should succeed and update the astroid cache
            importfrom_node = astroid.extract_node("from cache import a")
            importfrom_node.do_import_module(importfrom_node.modname)
        finally:
            sys.path = orig_sys_path

        astroid.MANAGER.clear_cache()

        importfrom_node = astroid.extract_node("from cache import a")
        # Try import from again after clear cache, this should raise an error
        with self.assertRaises(AstroidBuildingError):
            importfrom_node.do_import_module(importfrom_node.modname)

    def test_brain_plugins_reloaded_after_clearing_cache(self) -> None:
        astroid.MANAGER.clear_cache()
        format_call = astroid.extract_node("''.format()")
        inferred = next(format_call.infer())
        self.assertIsInstance(inferred, Const)

    def test_builtins_inference_after_clearing_cache(self) -> None:
        astroid.MANAGER.clear_cache()
        isinstance_call = astroid.extract_node("isinstance(1, int)")
        inferred = next(isinstance_call.infer())
        self.assertIs(inferred.value, True)

    def test_builtins_inference_after_clearing_cache_manually(self) -> None:
        # Not recommended to manipulate this, so we detect it and call clear_cache() instead
        astroid.MANAGER.brain["astroid_cache"].clear()
        isinstance_call = astroid.extract_node("isinstance(1, int)")
        inferred = next(isinstance_call.infer())
        self.assertIs(inferred.value, True)



# Licensed under the LGPL: https://www.gnu.org/licenses/old-licenses/lgpl-2.1.en.html
# For details: https://github.com/pylint-dev/astroid/blob/main/LICENSE
# Copyright (c) https://github.com/pylint-dev/astroid/blob/main/CONTRIBUTORS.txt

"""Tests for modules in the stdlib."""

from astroid import nodes
from astroid.builder import _extract_single_node


class TestSys:
    """Tests for the sys module."""

    def test_sys_builtin_module_names(self) -> None:
        """Test that we can gather the elements of a living tuple object."""
        node = _extract_single_node(
            """
        import sys
        sys.builtin_module_names
        """
        )
        inferred = list(node.infer())
        assert len(inferred) == 1
        assert isinstance(inferred[0], nodes.Tuple)
        assert inferred[0].elts

    def test_sys_modules(self) -> None:
        """Test that we can gather the items of a living dict object."""
        node = _extract_single_node(
            """
        import sys
        sys.modules
        """
        )
        inferred = list(node.infer())
        assert len(inferred) == 1
        assert isinstance(inferred[0], nodes.Dict)
        assert inferred[0].items



# Licensed under the LGPL: https://www.gnu.org/licenses/old-licenses/lgpl-2.1.en.html
# For details: https://github.com/pylint-dev/astroid/blob/main/LICENSE
# Copyright (c) https://github.com/pylint-dev/astroid/blob/main/CONTRIBUTORS.txt

import pytest

from astroid import bases, builder, nodes
from astroid.exceptions import InferenceError


def test_infer_typevar() -> None:
    """
    Regression test for: https://github.com/pylint-dev/pylint/issues/8802

    Test that an inferred `typing.TypeVar()` call produces a `nodes.ClassDef`
    node.
    """
    call_node = builder.extract_node(
        """
    from typing import TypeVar
    TypeVar('My.Type')
    """
    )
    with pytest.raises(InferenceError):
        call_node.inferred()


class TestTypingAlias:
    def test_infer_typing_alias(self) -> None:
        """
        Test that _alias() calls can be inferred.
        """
        node = builder.extract_node(
            """
            from typing import _alias
            x = _alias(int, float)
            """
        )
        assert isinstance(node, nodes.Assign)
        assert isinstance(node.value, nodes.Call)
        inferred = next(node.value.infer())
        assert isinstance(inferred, nodes.ClassDef)
        assert len(inferred.bases) == 1
        assert inferred.bases[0].name == "int"

    @pytest.mark.parametrize(
        "alias_args",
        [
            "",  # two missing arguments
            "int",  # one missing argument
            "int, float, tuple",  # one additional argument
        ],
    )
    def test_infer_typing_alias_incorrect_number_of_arguments(
        self, alias_args: str
    ) -> None:
        """
        Regression test for: https://github.com/pylint-dev/astroid/issues/2513

        Test that _alias() calls with the incorrect number of arguments can be inferred.
        """
        node = builder.extract_node(
            f"""
            from typing import _alias
            x = _alias({alias_args})
            """
        )
        assert isinstance(node, nodes.Assign)
        assert isinstance(node.value, nodes.Call)
        inferred = next(node.value.infer())
        assert isinstance(inferred, bases.Instance)
        assert inferred.name == "_SpecialGenericAlias"



"""This file contains Finders automatically generated by setuptools for a package installed
in editable mode via custom import hooks. It's generated here:
https://github.com/pypa/setuptools/blob/c34b82735c1a9c8707bea00705ae2f621bf4c24d/setuptools/command/editable_wheel.py#L732-L801
"""
import sys
from importlib.machinery import ModuleSpec
from importlib.machinery import all_suffixes as module_suffixes
from importlib.util import spec_from_file_location
from itertools import chain
from pathlib import Path

MAPPING = {"example": Path(__file__).parent.resolve() / "example"}
NAMESPACES = {}
PATH_PLACEHOLDER = "__editable__.example-0.1.0.finder" + ".__path_hook__"


class _EditableFinder:  # MetaPathFinder
    @classmethod
    def find_spec(cls, fullname, path=None, target=None):
        for pkg, pkg_path in reversed(list(MAPPING.items())):
            if fullname == pkg or fullname.startswith(f"{pkg}."):
                rest = fullname.replace(pkg, "", 1).strip(".").split(".")
                return cls._find_spec(fullname, Path(pkg_path, *rest))

        return None

    @classmethod
    def _find_spec(cls, fullname, candidate_path):
        init = candidate_path / "__init__.py"
        candidates = (candidate_path.with_suffix(x) for x in module_suffixes())
        for candidate in chain([init], candidates):
            if candidate.exists():
                return spec_from_file_location(fullname, candidate)


class _EditableNamespaceFinder:  # PathEntryFinder
    @classmethod
    def _path_hook(cls, path):
        if path == PATH_PLACEHOLDER:
            return cls
        raise ImportError

    @classmethod
    def _paths(cls, fullname):
        # Ensure __path__ is not empty for the spec to be considered a namespace.
        return NAMESPACES[fullname] or MAPPING.get(fullname) or [PATH_PLACEHOLDER]

    @classmethod
    def find_spec(cls, fullname, target=None):
        if fullname in NAMESPACES:
            spec = ModuleSpec(fullname, None, is_package=True)
            spec.submodule_search_locations = cls._paths(fullname)
            return spec
        return None

    @classmethod
    def find_module(cls, fullname):
        return None


def install():
    if not any(finder == _EditableFinder for finder in sys.meta_path):
        sys.meta_path.append(_EditableFinder)

    if not NAMESPACES:
        return

    if not any(hook == _EditableNamespaceFinder._path_hook for hook in sys.path_hooks):
        # PathEntryFinder is needed to create NamespaceSpec without private APIS
        sys.path_hooks.append(_EditableNamespaceFinder._path_hook)
    if PATH_PLACEHOLDER not in sys.path:
        sys.path.append(PATH_PLACEHOLDER)  # Used just to trigger the path hook



# Licensed under the LGPL: https://www.gnu.org/licenses/old-licenses/lgpl-2.1.en.html
# For details: https://github.com/pylint-dev/astroid/blob/main/LICENSE
# Copyright (c) https://github.com/pylint-dev/astroid/blob/main/CONTRIBUTORS.txt

try:
    import regex  # type: ignore[import]

    HAS_REGEX = True
except ImportError:
    HAS_REGEX = False

import pytest

from astroid import MANAGER, builder, nodes


@pytest.mark.skipif(not HAS_REGEX, reason="This test requires the regex library.")
class TestRegexBrain:
    def test_regex_flags(self) -> None:
        """Test that we have all regex enum flags in the brain."""
        names = [name for name in dir(regex) if name.isupper()]
        re_ast = MANAGER.ast_from_module_name("regex")
        for name in names:
            assert name in re_ast
            assert next(re_ast[name].infer()).value == getattr(regex, name)

    @pytest.mark.xfail(
        reason="Started failing on main, but no one reproduced locally yet"
    )
    def test_regex_pattern_and_match_subscriptable(self):
        """Test regex.Pattern and regex.Match are subscriptable in PY39+."""
        node1 = builder.extract_node(
            """
        import regex
        regex.Pattern[str]
        """
        )
        inferred1 = next(node1.infer())
        assert isinstance(inferred1, nodes.ClassDef)
        assert isinstance(inferred1.getattr("__class_getitem__")[0], nodes.FunctionDef)

        node2 = builder.extract_node(
            """
        import regex
        regex.Match[str]
        """
        )
        inferred2 = next(node2.infer())
        assert isinstance(inferred2, nodes.ClassDef)
        assert isinstance(inferred2.getattr("__class_getitem__")[0], nodes.FunctionDef)



# Licensed under the LGPL: https://www.gnu.org/licenses/old-licenses/lgpl-2.1.en.html
# For details: https://github.com/pylint-dev/astroid/blob/main/LICENSE
# Copyright (c) https://github.com/pylint-dev/astroid/blob/main/CONTRIBUTORS.txt

from __future__ import annotations

import unittest

from astroid import bases, builder, nodes, objects, util
from astroid.exceptions import AttributeInferenceError, InferenceError, SuperError
from astroid.objects import Super


class ObjectsTest(unittest.TestCase):
    def test_frozenset(self) -> None:
        node = builder.extract_node(
            """
        frozenset({1: 2, 2: 3}) #@
        """
        )
        inferred = next(node.infer())
        self.assertIsInstance(inferred, objects.FrozenSet)

        self.assertEqual(inferred.pytype(), "builtins.frozenset")

        itered = inferred.itered()
        self.assertEqual(len(itered), 2)
        self.assertIsInstance(itered[0], nodes.Const)
        self.assertEqual([const.value for const in itered], [1, 2])

        proxied = inferred._proxied
        self.assertEqual(inferred.qname(), "builtins.frozenset")
        self.assertIsInstance(proxied, nodes.ClassDef)

    def test_lookup_regression_slots(self) -> None:
        """Regression test for attr__new__ of ObjectModel.

        ObjectModel._instance is not always an bases.Instance, so we can't
        rely on the ._proxied attribute of an Instance.
        """

        node = builder.extract_node(
            """
        class ClassHavingUnknownAncestors(Unknown):
            __slots__ = ["yo"]

            def test(self):
                self.not_yo = 42
        """
        )
        assert node.getattr("__new__")


class SuperTests(unittest.TestCase):
    def test_inferring_super_outside_methods(self) -> None:
        ast_nodes = builder.extract_node(
            """
        class Module(object):
            pass
        class StaticMethod(object):
            @staticmethod
            def static():
                # valid, but we don't bother with it.
                return super(StaticMethod, StaticMethod) #@
        # super outside methods aren't inferred
        super(Module, Module) #@
        # no argument super is not recognised outside methods as well.
        super() #@
        """
        )
        assert isinstance(ast_nodes, list)
        in_static = next(ast_nodes[0].value.infer())
        self.assertIsInstance(in_static, bases.Instance)
        self.assertEqual(in_static.qname(), "builtins.super")

        module_level = next(ast_nodes[1].infer())
        self.assertIsInstance(module_level, bases.Instance)
        self.assertEqual(in_static.qname(), "builtins.super")

        no_arguments = next(ast_nodes[2].infer())
        self.assertIsInstance(no_arguments, bases.Instance)
        self.assertEqual(no_arguments.qname(), "builtins.super")

    def test_inferring_unbound_super_doesnt_work(self) -> None:
        node = builder.extract_node(
            """
        class Test(object):
            def __init__(self):
                super(Test) #@
        """
        )
        unbounded = next(node.infer())
        self.assertIsInstance(unbounded, bases.Instance)
        self.assertEqual(unbounded.qname(), "builtins.super")

    def test_use_default_inference_on_not_inferring_args(self) -> None:
        ast_nodes = builder.extract_node(
            """
        class Test(object):
            def __init__(self):
                super(Lala, self) #@
                super(Test, lala) #@
        """
        )
        assert isinstance(ast_nodes, list)
        first = next(ast_nodes[0].infer())
        self.assertIsInstance(first, bases.Instance)
        self.assertEqual(first.qname(), "builtins.super")

        second = next(ast_nodes[1].infer())
        self.assertIsInstance(second, bases.Instance)
        self.assertEqual(second.qname(), "builtins.super")

    def test_no_arguments_super(self) -> None:
        ast_nodes = builder.extract_node(
            """
        class First(object): pass
        class Second(First):
            def test(self):
                super() #@
            @classmethod
            def test_classmethod(cls):
                super() #@
        """
        )
        assert isinstance(ast_nodes, list)
        first = next(ast_nodes[0].infer())
        self.assertIsInstance(first, objects.Super)
        self.assertIsInstance(first.type, bases.Instance)
        self.assertEqual(first.type.name, "Second")
        self.assertIsInstance(first.mro_pointer, nodes.ClassDef)
        self.assertEqual(first.mro_pointer.name, "Second")

        second = next(ast_nodes[1].infer())
        self.assertIsInstance(second, objects.Super)
        self.assertIsInstance(second.type, nodes.ClassDef)
        self.assertEqual(second.type.name, "Second")
        self.assertIsInstance(second.mro_pointer, nodes.ClassDef)
        self.assertEqual(second.mro_pointer.name, "Second")

    def test_super_simple_cases(self) -> None:
        ast_nodes = builder.extract_node(
            """
        class First(object): pass
        class Second(First): pass
        class Third(First):
            def test(self):
                super(Third, self) #@
                super(Second, self) #@

                # mro position and the type
                super(Third, Third) #@
                super(Third, Second) #@
                super(Fourth, Fourth) #@

        class Fourth(Third):
            pass
        """
        )

        # .type is the object which provides the mro.
        # .mro_pointer is the position in the mro from where
        # the lookup should be done.

        # super(Third, self)
        assert isinstance(ast_nodes, list)
        first = next(ast_nodes[0].infer())
        self.assertIsInstance(first, objects.Super)
        self.assertIsInstance(first.type, bases.Instance)
        self.assertEqual(first.type.name, "Third")
        self.assertIsInstance(first.mro_pointer, nodes.ClassDef)
        self.assertEqual(first.mro_pointer.name, "Third")

        # super(Second, self)
        second = next(ast_nodes[1].infer())
        self.assertIsInstance(second, objects.Super)
        self.assertIsInstance(second.type, bases.Instance)
        self.assertEqual(second.type.name, "Third")
        self.assertIsInstance(first.mro_pointer, nodes.ClassDef)
        self.assertEqual(second.mro_pointer.name, "Second")

        # super(Third, Third)
        third = next(ast_nodes[2].infer())
        self.assertIsInstance(third, objects.Super)
        self.assertIsInstance(third.type, nodes.ClassDef)
        self.assertEqual(third.type.name, "Third")
        self.assertIsInstance(third.mro_pointer, nodes.ClassDef)
        self.assertEqual(third.mro_pointer.name, "Third")

        # super(Third, second)
        fourth = next(ast_nodes[3].infer())
        self.assertIsInstance(fourth, objects.Super)
        self.assertIsInstance(fourth.type, nodes.ClassDef)
        self.assertEqual(fourth.type.name, "Second")
        self.assertIsInstance(fourth.mro_pointer, nodes.ClassDef)
        self.assertEqual(fourth.mro_pointer.name, "Third")

        # Super(Fourth, Fourth)
        fifth = next(ast_nodes[4].infer())
        self.assertIsInstance(fifth, objects.Super)
        self.assertIsInstance(fifth.type, nodes.ClassDef)
        self.assertEqual(fifth.type.name, "Fourth")
        self.assertIsInstance(fifth.mro_pointer, nodes.ClassDef)
        self.assertEqual(fifth.mro_pointer.name, "Fourth")

    def test_super_infer(self) -> None:
        node = builder.extract_node(
            """
        class Super(object):
            def __init__(self):
                super(Super, self) #@
        """
        )
        inferred = next(node.infer())
        self.assertIsInstance(inferred, objects.Super)
        reinferred = next(inferred.infer())
        self.assertIsInstance(reinferred, objects.Super)
        self.assertIs(inferred, reinferred)

    def test_inferring_invalid_supers(self) -> None:
        ast_nodes = builder.extract_node(
            """
        class Super(object):
            def __init__(self):
                # MRO pointer is not a type
                super(1, self) #@
                # MRO type is not a subtype
                super(Super, 1) #@
                # self is not a subtype of Bupper
                super(Bupper, self) #@
        class Bupper(Super):
            pass
        """
        )
        assert isinstance(ast_nodes, list)
        first = next(ast_nodes[0].infer())
        self.assertIsInstance(first, objects.Super)
        with self.assertRaises(SuperError) as cm:
            first.super_mro()
        self.assertIsInstance(cm.exception.super_.mro_pointer, nodes.Const)
        self.assertEqual(cm.exception.super_.mro_pointer.value, 1)
        for node, invalid_type in zip(ast_nodes[1:], (nodes.Const, bases.Instance)):
            inferred = next(node.infer())
            self.assertIsInstance(inferred, objects.Super, node)
            with self.assertRaises(SuperError) as cm:
                inferred.super_mro()
            self.assertIsInstance(cm.exception.super_.type, invalid_type)

    def test_proxied(self) -> None:
        node = builder.extract_node(
            """
        class Super(object):
            def __init__(self):
                super(Super, self) #@
        """
        )
        inferred = next(node.infer())
        proxied = inferred._proxied
        self.assertEqual(proxied.qname(), "builtins.super")
        self.assertIsInstance(proxied, nodes.ClassDef)

    def test_super_bound_model(self) -> None:
        ast_nodes = builder.extract_node(
            """
        class First(object):
            def method(self):
                pass
            @classmethod
            def class_method(cls):
                pass
        class Super_Type_Type(First):
            def method(self):
                super(Super_Type_Type, Super_Type_Type).method #@
                super(Super_Type_Type, Super_Type_Type).class_method #@
            @classmethod
            def class_method(cls):
                super(Super_Type_Type, Super_Type_Type).method #@
                super(Super_Type_Type, Super_Type_Type).class_method #@

        class Super_Type_Object(First):
            def method(self):
                super(Super_Type_Object, self).method #@
                super(Super_Type_Object, self).class_method #@
        """
        )
        # Super(type, type) is the same for both functions and classmethods.
        assert isinstance(ast_nodes, list)
        first = next(ast_nodes[0].infer())
        self.assertIsInstance(first, nodes.FunctionDef)
        self.assertEqual(first.name, "method")

        second = next(ast_nodes[1].infer())
        self.assertIsInstance(second, bases.BoundMethod)
        self.assertEqual(second.bound.name, "First")
        self.assertEqual(second.type, "classmethod")

        third = next(ast_nodes[2].infer())
        self.assertIsInstance(third, nodes.FunctionDef)
        self.assertEqual(third.name, "method")

        fourth = next(ast_nodes[3].infer())
        self.assertIsInstance(fourth, bases.BoundMethod)
        self.assertEqual(fourth.bound.name, "First")
        self.assertEqual(fourth.type, "classmethod")

        # Super(type, obj) can lead to different attribute bindings
        # depending on the type of the place where super was called.
        fifth = next(ast_nodes[4].infer())
        self.assertIsInstance(fifth, bases.BoundMethod)
        self.assertEqual(fifth.bound.name, "First")
        self.assertEqual(fifth.type, "method")

        sixth = next(ast_nodes[5].infer())
        self.assertIsInstance(sixth, bases.BoundMethod)
        self.assertEqual(sixth.bound.name, "First")
        self.assertEqual(sixth.type, "classmethod")

    def test_super_getattr_single_inheritance(self) -> None:
        ast_nodes = builder.extract_node(
            """
        class First(object):
            def test(self): pass
        class Second(First):
            def test2(self): pass
        class Third(Second):
            test3 = 42
            def __init__(self):
                super(Third, self).test2 #@
                super(Third, self).test #@
                # test3 is local, no MRO lookup is done.
                super(Third, self).test3 #@
                super(Third, self) #@

                # Unbounds.
                super(Third, Third).test2 #@
                super(Third, Third).test #@

        """
        )
        assert isinstance(ast_nodes, list)
        first = next(ast_nodes[0].infer())
        self.assertIsInstance(first, bases.BoundMethod)
        self.assertEqual(first.bound.name, "Second")

        second = next(ast_nodes[1].infer())
        self.assertIsInstance(second, bases.BoundMethod)
        self.assertEqual(second.bound.name, "First")

        with self.assertRaises(InferenceError):
            next(ast_nodes[2].infer())
        fourth = next(ast_nodes[3].infer())
        with self.assertRaises(AttributeInferenceError):
            fourth.getattr("test3")
        with self.assertRaises(AttributeInferenceError):
            next(fourth.igetattr("test3"))

        first_unbound = next(ast_nodes[4].infer())
        self.assertIsInstance(first_unbound, nodes.FunctionDef)
        self.assertEqual(first_unbound.name, "test2")
        self.assertEqual(first_unbound.parent.name, "Second")

        second_unbound = next(ast_nodes[5].infer())
        self.assertIsInstance(second_unbound, nodes.FunctionDef)
        self.assertEqual(second_unbound.name, "test")
        self.assertEqual(second_unbound.parent.name, "First")

    def test_super_invalid_mro(self) -> None:
        node = builder.extract_node(
            """
        class A(object):
           test = 42
        class Super(A, A):
           def __init__(self):
               super(Super, self) #@
        """
        )
        inferred = next(node.infer())
        with self.assertRaises(AttributeInferenceError):
            next(inferred.getattr("test"))

    def test_super_complex_mro(self) -> None:
        ast_nodes = builder.extract_node(
            """
        class A(object):
            def spam(self): return "A"
            def foo(self): return "A"
            @staticmethod
            def static(self): pass
        class B(A):
            def boo(self): return "B"
            def spam(self): return "B"
        class C(A):
            def boo(self): return "C"
        class E(C, B):
            def __init__(self):
                super(E, self).boo #@
                super(C, self).boo #@
                super(E, self).spam #@
                super(E, self).foo #@
                super(E, self).static #@
        """
        )
        assert isinstance(ast_nodes, list)
        first = next(ast_nodes[0].infer())
        self.assertIsInstance(first, bases.BoundMethod)
        self.assertEqual(first.bound.name, "C")
        second = next(ast_nodes[1].infer())
        self.assertIsInstance(second, bases.BoundMethod)
        self.assertEqual(second.bound.name, "B")
        third = next(ast_nodes[2].infer())
        self.assertIsInstance(third, bases.BoundMethod)
        self.assertEqual(third.bound.name, "B")
        fourth = next(ast_nodes[3].infer())
        self.assertEqual(fourth.bound.name, "A")
        static = next(ast_nodes[4].infer())
        self.assertIsInstance(static, nodes.FunctionDef)
        self.assertEqual(static.parent.scope().name, "A")

    def test_super_data_model(self) -> None:
        ast_nodes = builder.extract_node(
            """
        class X(object): pass
        class A(X):
            def __init__(self):
                super(A, self) #@
                super(A, A) #@
                super(X, A) #@
        """
        )
        assert isinstance(ast_nodes, list)
        first = next(ast_nodes[0].infer())
        thisclass = first.getattr("__thisclass__")[0]
        self.assertIsInstance(thisclass, nodes.ClassDef)
        self.assertEqual(thisclass.name, "A")
        selfclass = first.getattr("__self_class__")[0]
        self.assertIsInstance(selfclass, nodes.ClassDef)
        self.assertEqual(selfclass.name, "A")
        self_ = first.getattr("__self__")[0]
        self.assertIsInstance(self_, bases.Instance)
        self.assertEqual(self_.name, "A")
        cls = first.getattr("__class__")[0]
        self.assertEqual(cls, first._proxied)

        second = next(ast_nodes[1].infer())
        thisclass = second.getattr("__thisclass__")[0]
        self.assertEqual(thisclass.name, "A")
        self_ = second.getattr("__self__")[0]
        self.assertIsInstance(self_, nodes.ClassDef)
        self.assertEqual(self_.name, "A")

        third = next(ast_nodes[2].infer())
        thisclass = third.getattr("__thisclass__")[0]
        self.assertEqual(thisclass.name, "X")
        selfclass = third.getattr("__self_class__")[0]
        self.assertEqual(selfclass.name, "A")

    def assertEqualMro(self, klass: Super, expected_mro: list[str]) -> None:
        self.assertEqual([member.name for member in klass.super_mro()], expected_mro)

    def test_super_mro(self) -> None:
        ast_nodes = builder.extract_node(
            """
        class A(object): pass
        class B(A): pass
        class C(A): pass
        class E(C, B):
            def __init__(self):
                super(E, self) #@
                super(C, self) #@
                super(B, self) #@

                super(B, 1) #@
                super(1, B) #@
        """
        )
        assert isinstance(ast_nodes, list)
        first = next(ast_nodes[0].infer())
        self.assertEqualMro(first, ["C", "B", "A", "object"])
        second = next(ast_nodes[1].infer())
        self.assertEqualMro(second, ["B", "A", "object"])
        third = next(ast_nodes[2].infer())
        self.assertEqualMro(third, ["A", "object"])

        fourth = next(ast_nodes[3].infer())
        with self.assertRaises(SuperError):
            fourth.super_mro()
        fifth = next(ast_nodes[4].infer())
        with self.assertRaises(SuperError):
            fifth.super_mro()

    def test_super_yes_objects(self) -> None:
        ast_nodes = builder.extract_node(
            """
        from collections import Missing
        class A(object):
            def __init__(self):
                super(Missing, self) #@
                super(A, Missing) #@
        """
        )
        assert isinstance(ast_nodes, list)
        first = next(ast_nodes[0].infer())
        self.assertIsInstance(first, bases.Instance)
        second = next(ast_nodes[1].infer())
        self.assertIsInstance(second, bases.Instance)

    def test_super_invalid_types(self) -> None:
        node = builder.extract_node(
            """
        import collections
        class A(object):
            def __init__(self):
                super(A, collections) #@
        """
        )
        inferred = next(node.infer())
        with self.assertRaises(SuperError):
            inferred.super_mro()
        with self.assertRaises(SuperError):
            inferred.super_mro()

    def test_super_properties(self) -> None:
        node = builder.extract_node(
            """
        class Foo(object):
            @property
            def dict(self):
                return 42

        class Bar(Foo):
            @property
            def dict(self):
                return super(Bar, self).dict

        Bar().dict
        """
        )
        inferred = next(node.infer())
        self.assertIsInstance(inferred, nodes.Const)
        self.assertEqual(inferred.value, 42)

    def test_super_qname(self) -> None:
        """Make sure a Super object generates a qname
        equivalent to super.__qname__.
        """
        # See issue 533
        code = """
        class C:
           def foo(self): return super()
        C().foo() #@
        """
        super_obj = next(builder.extract_node(code).infer())
        self.assertEqual(super_obj.qname(), "super")

    def test_super_new_call(self) -> None:
        """Test that __new__ returns an object or node and not a (Un)BoundMethod."""
        new_call_result: nodes.Name = builder.extract_node(
            """
        import enum
        class ChoicesMeta(enum.EnumMeta):
            def __new__(metacls, classname, bases, classdict, **kwds):
                cls = super().__new__(metacls, "str", (enum.Enum,), enum._EnumDict(), **kwargs)
                cls #@
        """
        )
        inferred = list(new_call_result.infer())
        assert all(
            isinstance(i, (nodes.NodeNG, type(util.Uninferable))) for i in inferred
        )

    def test_super_init_call(self) -> None:
        """Test that __init__ is still callable."""
        init_node: nodes.Attribute = builder.extract_node(
            """
        class SuperUsingClass:
            @staticmethod
            def test():
                super(object, 1).__new__ #@
                super(object, 1).__init__ #@
        class A:
            pass
        A().__new__ #@
        A().__init__ #@
        """
        )
        assert isinstance(next(init_node[0].infer()), bases.BoundMethod)
        assert isinstance(next(init_node[1].infer()), bases.BoundMethod)
        assert isinstance(next(init_node[2].infer()), bases.BoundMethod)
        assert isinstance(next(init_node[3].infer()), bases.BoundMethod)



LONG_CHAINED_METHOD_CALL = """
from a import b

(
    b.builder('name')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .add('name', value='value')
    .Build()
)"""



# Licensed under the LGPL: https://www.gnu.org/licenses/old-licenses/lgpl-2.1.en.html
# For details: https://github.com/pylint-dev/astroid/blob/main/LICENSE
# Copyright (c) https://github.com/pylint-dev/astroid/blob/main/CONTRIBUTORS.txt

from __future__ import annotations

import unittest

import astroid
from astroid import builder
from astroid.bases import Instance


class ThreadingBrainTest(unittest.TestCase):
    def test_lock(self) -> None:
        lock_instance = builder.extract_node(
            """
        import threading
        threading.Lock()
        """
        )
        inferred = next(lock_instance.infer())
        self.assert_is_valid_lock(inferred)

        acquire_method = inferred.getattr("acquire")[0]
        parameters = [param.name for param in acquire_method.args.args[1:]]
        assert parameters == ["blocking", "timeout"]

        assert inferred.getattr("locked")

    def test_rlock(self) -> None:
        self._test_lock_object("RLock")

    def test_semaphore(self) -> None:
        self._test_lock_object("Semaphore")

    def test_boundedsemaphore(self) -> None:
        self._test_lock_object("BoundedSemaphore")

    def _test_lock_object(self, object_name: str) -> None:
        lock_instance = builder.extract_node(
            f"""
        import threading
        threading.{object_name}()
        """
        )
        inferred = next(lock_instance.infer())
        self.assert_is_valid_lock(inferred)

    def assert_is_valid_lock(self, inferred: Instance) -> None:
        self.assertIsInstance(inferred, astroid.Instance)
        self.assertEqual(inferred.root().name, "threading")
        for method in ("acquire", "release", "__enter__", "__exit__"):
            self.assertIsInstance(next(inferred.igetattr(method)), astroid.BoundMethod)



assert not not True == True
assert (not False or True) == True
assert True or False and True
assert (True or False) and True

assert True is not (False is True) == False
assert True is (not False is True == False)

assert 1 + 2 + 3 == 6
assert 5 - 4 + 3 == 4
assert 4 - 5 - 6 == -7
assert 7 - (8 - 9) == 8
assert 2**3**4 == 2**81
assert (2**3)**4 == 8**4

assert 1 + 2 if (0.5 if True else 0.2) else 1 if True else 2 == 3
assert (0 if True else 1) if False else 2 == 2
assert lambda x: x if (0 if False else 0) else 0 if False else 0
assert (lambda x: x) if (0 if True else 0.2) else 1 if True else 2

assert ('1' + '2').replace('1', '3') == '32'
assert (lambda x: x)(1) == 1
assert ([0] + [1])[1] == 1
assert (lambda x: lambda: x + 1)(2)() == 3

f = lambda x, y, z: y(x, z)
assert f(1, lambda x, y: x + y[1], (2, 3)) == 4



# Licensed under the LGPL: https://www.gnu.org/licenses/old-licenses/lgpl-2.1.en.html
# For details: https://github.com/pylint-dev/astroid/blob/main/LICENSE
# Copyright (c) https://github.com/pylint-dev/astroid/blob/main/CONTRIBUTORS.txt

"""Astroid hooks for numpy.core.fromnumeric module."""
from astroid.brain.helpers import register_module_extender
from astroid.builder import parse
from astroid.manager import AstroidManager


def numpy_core_fromnumeric_transform():
    return parse(
        """
    def sum(a, axis=None, dtype=None, out=None, keepdims=None, initial=None):
        return numpy.ndarray([0, 0])
    """
    )


def register(manager: AstroidManager) -> None:
    register_module_extender(
        manager, "numpy.core.fromnumeric", numpy_core_fromnumeric_transform
    )



# Licensed under the LGPL: https://www.gnu.org/licenses/old-licenses/lgpl-2.1.en.html
# For details: https://github.com/pylint-dev/astroid/blob/main/LICENSE
# Copyright (c) https://github.com/pylint-dev/astroid/blob/main/CONTRIBUTORS.txt

import unittest

try:
    import numpy  # pylint: disable=unused-import

    HAS_NUMPY = True
except ImportError:
    HAS_NUMPY = False

from astroid import bases, builder, nodes


@unittest.skipUnless(HAS_NUMPY, "This test requires the numpy library.")
class NumpyBrainCoreUmathTest(unittest.TestCase):
    """Test of all members of numpy.core.umath module."""

    one_arg_ufunc = (
        "arccos",
        "arccosh",
        "arcsin",
        "arcsinh",
        "arctan",
        "arctanh",
        "cbrt",
        "conj",
        "conjugate",
        "cosh",
        "deg2rad",
        "degrees",
        "exp2",
        "expm1",
        "fabs",
        "frexp",
        "isfinite",
        "isinf",
        "log",
        "log1p",
        "log2",
        "logical_not",
        "modf",
        "negative",
        "positive",
        "rad2deg",
        "radians",
        "reciprocal",
        "rint",
        "sign",
        "signbit",
        "spacing",
        "square",
        "tan",
        "tanh",
        "trunc",
    )

    two_args_ufunc = (
        "add",
        "bitwise_and",
        "bitwise_or",
        "bitwise_xor",
        "copysign",
        "divide",
        "divmod",
        "equal",
        "float_power",
        "floor_divide",
        "fmax",
        "fmin",
        "fmod",
        "gcd",
        "greater",
        "heaviside",
        "hypot",
        "lcm",
        "ldexp",
        "left_shift",
        "less",
        "logaddexp",
        "logaddexp2",
        "logical_and",
        "logical_or",
        "logical_xor",
        "maximum",
        "minimum",
        "multiply",
        "nextafter",
        "not_equal",
        "power",
        "remainder",
        "right_shift",
        "subtract",
        "true_divide",
    )

    all_ufunc = one_arg_ufunc + two_args_ufunc

    constants = ("e", "euler_gamma")

    def _inferred_numpy_attribute(self, func_name):
        node = builder.extract_node(
            f"""
        import numpy.core.umath as tested_module
        func = tested_module.{func_name:s}
        func"""
        )
        return next(node.infer())

    def test_numpy_core_umath_constants(self):
        """Test that constants have Const type."""
        for const in self.constants:
            with self.subTest(const=const):
                inferred = self._inferred_numpy_attribute(const)
                self.assertIsInstance(inferred, nodes.Const)

    def test_numpy_core_umath_constants_values(self):
        """Test the values of the constants."""
        exact_values = {"e": 2.718281828459045, "euler_gamma": 0.5772156649015329}
        for const in self.constants:
            with self.subTest(const=const):
                inferred = self._inferred_numpy_attribute(const)
                self.assertEqual(inferred.value, exact_values[const])

    def test_numpy_core_umath_functions(self):
        """Test that functions have FunctionDef type."""
        for func in self.all_ufunc:
            with self.subTest(func=func):
                inferred = self._inferred_numpy_attribute(func)
                self.assertIsInstance(inferred, bases.Instance)

    def test_numpy_core_umath_functions_one_arg(self):
        """Test the arguments names of functions."""
        exact_arg_names = [
            "self",
            "x",
            "out",
            "where",
            "casting",
            "order",
            "dtype",
            "subok",
        ]
        for func in self.one_arg_ufunc:
            with self.subTest(func=func):
                inferred = self._inferred_numpy_attribute(func)
                self.assertEqual(
                    inferred.getattr("__call__")[0].argnames(), exact_arg_names
                )

    def test_numpy_core_umath_functions_two_args(self):
        """Test the arguments names of functions."""
        exact_arg_names = [
            "self",
            "x1",
            "x2",
            "out",
            "where",
            "casting",
            "order",
            "dtype",
            "subok",
        ]
        for func in self.two_args_ufunc:
            with self.subTest(func=func):
                inferred = self._inferred_numpy_attribute(func)
                self.assertEqual(
                    inferred.getattr("__call__")[0].argnames(), exact_arg_names
                )

    def test_numpy_core_umath_functions_kwargs_default_values(self):
        """Test the default values for keyword arguments."""
        exact_kwargs_default_values = [None, True, "same_kind", "K", None, True]
        for func in self.one_arg_ufunc + self.two_args_ufunc:
            with self.subTest(func=func):
                inferred = self._inferred_numpy_attribute(func)
                default_args_values = [
                    default.value
                    for default in inferred.getattr("__call__")[0].args.defaults
                ]
                self.assertEqual(default_args_values, exact_kwargs_default_values)

    def _inferred_numpy_func_call(self, func_name, *func_args):
        node = builder.extract_node(
            f"""
        import numpy as np
        func = np.{func_name:s}
        func()
        """
        )
        return node.infer()

    def test_numpy_core_umath_functions_return_type(self):
        """Test that functions which should return a ndarray do return it."""
        ndarray_returning_func = [
            f for f in self.all_ufunc if f not in ("frexp", "modf")
        ]
        for func_ in ndarray_returning_func:
            with self.subTest(typ=func_):
                inferred_values = list(self._inferred_numpy_func_call(func_))
                self.assertTrue(
                    len(inferred_values) == 1,
                    msg="Too much inferred values ({}) for {:s}".format(
                        inferred_values[-1].pytype(), func_
                    ),
                )
                self.assertTrue(
                    inferred_values[0].pytype() == ".ndarray",
                    msg=f"Illicit type for {func_:s} ({inferred_values[-1].pytype()})",
                )

    def test_numpy_core_umath_functions_return_type_tuple(self):
        """Test that functions which should return a pair of ndarray do return it."""
        ndarray_returning_func = ("frexp", "modf")

        for func_ in ndarray_returning_func:
            with self.subTest(typ=func_):
                inferred_values = list(self._inferred_numpy_func_call(func_))
                self.assertTrue(
                    len(inferred_values) == 1,
                    msg=f"Too much inferred values ({inferred_values}) for {func_:s}",
                )
                self.assertTrue(
                    inferred_values[-1].pytype() == "builtins.tuple",
                    msg=f"Illicit type for {func_:s} ({inferred_values[-1].pytype()})",
                )
                self.assertTrue(
                    len(inferred_values[0].elts) == 2,
                    msg=f"{func_} should return a pair of values. That's not the case.",
                )
                for array in inferred_values[-1].elts:
                    effective_infer = [m.pytype() for m in array.inferred()]
                    self.assertTrue(
                        ".ndarray" in effective_infer,
                        msg=(
                            f"Each item in the return of {func_} should be inferred"
                            f" as a ndarray and not as {effective_infer}"
                        ),
                    )



# Licensed under the LGPL: https://www.gnu.org/licenses/old-licenses/lgpl-2.1.en.html
# For details: https://github.com/pylint-dev/astroid/blob/main/LICENSE
# Copyright (c) https://github.com/pylint-dev/astroid/blob/main/CONTRIBUTORS.txt

from pathlib import Path

from contributors_txt import create_contributors_txt

CWD = Path(".").absolute()
ASTROID_BASE_DIRECTORY = Path(__file__).parent.parent.absolute()
ALIASES_FILE = (
    ASTROID_BASE_DIRECTORY / "script/.contributors_aliases.json"
).relative_to(CWD)
DEFAULT_CONTRIBUTOR_PATH = (ASTROID_BASE_DIRECTORY / "CONTRIBUTORS.txt").relative_to(
    CWD
)


def main():
    create_contributors_txt(
        aliases_file=ALIASES_FILE, output=DEFAULT_CONTRIBUTOR_PATH, verbose=True
    )


if __name__ == "__main__":
    main()



# Licensed under the LGPL: https://www.gnu.org/licenses/old-licenses/lgpl-2.1.en.html
# For details: https://github.com/pylint-dev/astroid/blob/main/LICENSE
# Copyright (c) https://github.com/pylint-dev/astroid/blob/main/CONTRIBUTORS.txt

from __future__ import annotations

import pytest

from astroid import builder, nodes

try:
    import numpy  # pylint: disable=unused-import

    HAS_NUMPY = True
except ImportError:
    HAS_NUMPY = False


def _inferred_numpy_func_call(func_name: str, *func_args: str) -> nodes.FunctionDef:
    node = builder.extract_node(
        f"""
    import numpy as np
    func = np.{func_name:s}
    func({','.join(func_args):s})
    """
    )
    return node.infer()


@pytest.mark.skipif(not HAS_NUMPY, reason="This test requires the numpy library.")
def test_numpy_function_calls_inferred_as_ndarray() -> None:
    """Test that calls to numpy functions are inferred as numpy.ndarray."""
    method = "einsum"
    inferred_values = list(
        _inferred_numpy_func_call(method, "ii, np.arange(25).reshape(5, 5)")
    )

    assert len(inferred_values) == 1, f"Too much inferred value for {method:s}"
    assert (
        inferred_values[-1].pytype() == ".ndarray"
    ), f"Illicit type for {method:s} ({inferred_values[-1].pytype()})"


@pytest.mark.skipif(not HAS_NUMPY, reason="This test requires the numpy library.")
def test_function_parameters() -> None:
    instance = builder.extract_node(
        """
    import numpy
    numpy.einsum #@
    """
    )
    actual_args = instance.inferred()[0].args

    assert actual_args.vararg == "operands"
    assert [arg.name for arg in actual_args.kwonlyargs] == ["out", "optimize"]
    assert actual_args.kwarg == "kwargs"



# Licensed under the LGPL: https://www.gnu.org/licenses/old-licenses/lgpl-2.1.en.html
# For details: https://github.com/pylint-dev/astroid/blob/main/LICENSE
# Copyright (c) https://github.com/pylint-dev/astroid/blob/main/CONTRIBUTORS.txt

"""Tests for the ssl brain."""

import pytest

from astroid import bases, nodes, parse
from astroid.const import PY312_PLUS


def test_ssl_brain() -> None:
    """Test ssl brain transform."""
    module = parse(
        """
    import ssl
    ssl.PROTOCOL_TLSv1
    ssl.VerifyMode
    ssl.TLSVersion
    ssl.VerifyMode.CERT_REQUIRED
    """
    )
    inferred_protocol = next(module.body[1].value.infer())
    assert isinstance(inferred_protocol, nodes.Const)

    inferred_verifymode = next(module.body[2].value.infer())
    assert isinstance(inferred_verifymode, nodes.ClassDef)
    assert inferred_verifymode.name == "VerifyMode"
    assert len(inferred_verifymode.bases) == 1

    # Check that VerifyMode correctly inherits from enum.IntEnum
    int_enum = next(inferred_verifymode.bases[0].infer())
    assert isinstance(int_enum, nodes.ClassDef)
    assert int_enum.name == "IntEnum"
    assert int_enum.parent.name == "enum"

    # TLSVersion is inferred from the main module, not from the brain
    inferred_tlsversion = next(module.body[3].value.infer())
    assert isinstance(inferred_tlsversion, nodes.ClassDef)
    assert inferred_tlsversion.name == "TLSVersion"

    # TLSVersion is inferred from the main module, not from the brain
    inferred_cert_required = next(module.body[4].value.infer())
    assert isinstance(inferred_cert_required, bases.Instance)
    assert inferred_cert_required._proxied.name == "CERT_REQUIRED"


@pytest.mark.skipif(not PY312_PLUS, reason="Uses new 3.12 constant")
def test_ssl_brain_py312() -> None:
    """Test ssl brain transform."""
    module = parse(
        """
    import ssl
    ssl.OP_LEGACY_SERVER_CONNECT
    ssl.Options.OP_LEGACY_SERVER_CONNECT
    """
    )

    inferred_constant = next(module.body[1].value.infer())
    assert isinstance(inferred_constant, nodes.Const)

    inferred_instance = next(module.body[2].value.infer())
    assert isinstance(inferred_instance, bases.Instance)
